import gfapy
from .alignment import Alignment

class Trace(list):
  """Trace alignment.

  A trace is a list of integers, each giving the number of characters
  in the second segment to align to the next ``TS`` characters in the first
  segment (where  ``TS``, the trace spacing, is either the default spacing
  given in the header line ``TS`` tag, or the the spacing given in the ``TS``
  tag on the line itself, where the trace alignment is used).

  Instances are usually created from their string representations, using the
  :class:`~gfapy.alignment.alignment.Alignment` factory class constructor.
  """

  def complement(self):
    """Computes the complement of the trace alignment.

    A complement operation (such as for CIGARs) cannot be defined
    for a trace, without computing the alignment. This is currently not
    available in gfapy.

    Returns:
      gfapy.AlignmentPlaceholder
    """
    return gfapy.AlignmentPlaceholder()

  def validate(self, ts = None, version = "gfa2"):
    """Validates the trace alignment

    Parameters:
      ts (int): Trace Spacing. If specified, it will be checked that all values
        are < **ts** (default: **None**, no check).
      version (str) : GFA version (must be 'gfa1' or 'gfa2')

    Raises:
      ~gfapy.error.TypeError: If the list contains non-integer values
      ~gfapy.error.ValueError: If the list contains values < 0 or > **ts**
      ~gfapy.error.VersionError: If the version is 'gfa1' or an invalid version
        string is provided
    """
    if version != "gfa2":
      if version == "gfa1":
        raise gfapy.VersionError("Traces are not compatible with GFA1")
      else:
        raise gfapy.VersionError("Version unknown: {}".format(repr(version)))
    for e in self:
      if not isinstance(e, int):
        raise gfapy.TypeError(
            ("Trace contains non-integer values ({0} found)\n" + "Content: {1}")
            .format(e, repr(self)))
      if e < 0:
        raise gfapy.ValueError(
            ("Trace contains value < 0 ({0} found)\n" + "Content: {1}")
            .format(e, repr(self)))
      if ts and e > ts:
        raise gfapy.ValueError(
            ("Trace contains value > TS ({0} found, TS = {2})\n" + "Content: {1}")
            .format(e, repr(self), ts))

  def __str__(self):
    if not self:
      return "*"
    else:
      return ",".join([str(v) for v in self])

  def __repr__(self):
    if not self:
      return 'gfapy.Trace([])'
    else:
      return "gfapy.Trace([{}])".format(str(self))

  @classmethod
  def _from_string(cls,string):
    try:
      return Trace([int(v) for v in string.split(",")])
    except:
      raise gfapy.FormatError("string does not encode"+
          " a valid trace alignment: {}".format(string))



import gfapy
import unittest

class TestAPIMultiplication(unittest.TestCase):

  def test_multiply_segment(self):
    gfa = gfapy.Gfa()
    gfa.append("H\tVN:Z:1.0")
    s = {"S\t0\t*\tRC:i:600",
         "S\t1\t*\tRC:i:6000",
         "S\t2\t*\tRC:i:60000"}
    l = "L\t1\t+\t2\t+\t12M"
    c = "C\t1\t+\t0\t+\t12\t12M"
    p = "P\t3\t2+,0-\t12M"
    for line in (list(s) + [l,c,p]): gfa.append(line)
    self.assertEqual(s, {str(x) for x in gfa.segments})
    self.assertEqual([l], [str(x) for x in gfa.dovetails if not x.virtual])
    self.assertEqual([c], [str(x) for x in gfa.containments])
    self.assertEqual([l], [str(x) for x in gfa.segment("1").end_relations("R", ["2", "L"])])
    self.assertEqual([c], [str(x) for x in gfa.segment("1").relations_to("0")])
    self.assertEqual(6000, gfa.segment("1").RC)
    gfa.multiply("1", 2)
    self.assertEqual([l], [str(x) for x in gfa.segment("1").end_relations("R", ["2", "L"])])
    self.assertEqual([c], [str(x) for x in gfa.segment("1").relations_to("0")])
    self.assertNotEqual([], [str(x) for x in gfa.segment("1*2").end_relations("R", ["2", "L"])])
    self.assertNotEqual([], [str(x) for x in gfa.segment("1*2").relations_to("0")])
    self.assertEqual(3000, gfa.segment("1").RC)
    self.assertEqual(3000, gfa.segment("1*2").RC)
    gfa.multiply("1*2", 3, copy_names=["6","7"])
    self.assertEqual([l], [str(x) for x in gfa.segment("1").end_relations("R", ["2", "L"])])
    self.assertNotEqual([], [str(x) for x in gfa.segment("1*2").end_relations("R", ["2", "L"])])
    self.assertNotEqual([], [str(x) for x in gfa.segment("6").end_relations("R", ["2", "L"])])
    self.assertNotEqual([], [str(x) for x in gfa.segment("7").end_relations("R", ["2", "L"])])
    self.assertNotEqual([], gfa.segment("1*2").relations_to("0"))
    self.assertNotEqual([], gfa.segment("6").relations_to("0"))
    self.assertNotEqual([], gfa.segment("7").relations_to("0"))
    self.assertEqual(3000, gfa.segment("1").RC)
    self.assertEqual(1000, gfa.segment("1*2").RC)
    self.assertEqual(1000, gfa.segment("6").RC)
    self.assertEqual(1000, gfa.segment("7").RC)

  def test_multiply_segment_copy_names(self):
    gfa = gfapy.Gfa(["H\tVN:Z:1.0",
           "S\t1\t*\tRC:i:600",
           "S\t1b\t*\tRC:i:6000",
           "S\t2\t*\tRC:i:60000",
           "S\t3\t*\tRC:i:60000"])
    gfa.multiply("2", 2)
    gfa.try_get_segment("2*2") # nothing raised
    gfa.multiply("2*2", 2)
    gfa.try_get_segment("2*3") # nothing raised
    gfa.multiply("2*2", 2, copy_names = ["x"])
    gfa.try_get_segment("x") # nothing raised

  def test_links_distribution_l1_m2(self):
    for sfx in ["gfa", "gfa2"]:
      g1 = gfapy.Gfa.from_file("tests/testdata/links_distri.l1.{}".format(sfx))
      g2 = gfapy.Gfa.from_file("tests/testdata/links_distri.l1.m2.{}".format(sfx))
      self.assertNotEqual(set(g2.segment_names),set(g1.segment_names))
      self.assertNotEqual(set([str(x) for x in g2.dovetails]),
                          set([str(x) for x in g1.dovetails]))
      g1.multiply("1", 2, extended=True)
      self.assertEqual(set(g2.segment_names),set(g1.segment_names))
      self.assertEqual(set([str(x) for x in g2.dovetails]),
                       set([str(x) for x in g1.dovetails]))

  def test_links_distribution_l2_m2(self):
    for sfx in ["gfa", "gfa2"]:
      g1 = gfapy.Gfa.from_file("tests/testdata/links_distri.l2.{}".format(sfx))
      g2 = gfapy.Gfa.from_file("tests/testdata/links_distri.l2.m2.{}".format(sfx))
      self.assertNotEqual(set(g2.segment_names),set(g1.segment_names))
      self.assertNotEqual(set([str(x) for x in g2.dovetails]),
                          set([str(x) for x in g1.dovetails]))
      g1.multiply("1", 2, extended=True)
      self.assertEqual(set(g2.segment_names),set(g1.segment_names))
      self.assertEqual(set([str(x) for x in g2.dovetails]),
                       set([str(x) for x in g1.dovetails]))

  def test_no_links_distribution_l2_m2(self):
    for sfx in ["gfa", "gfa2"]:
      g1 = gfapy.Gfa.from_file("tests/testdata/links_distri.l2.{}".format(sfx))
      g2 = gfapy.Gfa.from_file("tests/testdata/links_distri.l2.m2.no_ld.{}".format(sfx))
      self.assertNotEqual(set(g2.segment_names),set(g1.segment_names))
      self.assertNotEqual(set([str(x) for x in g2.dovetails]),
                          set([str(x) for x in g1.dovetails]))
      g1.multiply("1", 2, extended=True, distribute="off")
      self.assertEqual(set(g2.segment_names),set(g1.segment_names))
      self.assertEqual(set([str(x) for x in g2.dovetails]),
                       set([str(x) for x in g1.dovetails]))

  def test_links_distribution_l2_m3(self):
    for sfx in ["gfa", "gfa2"]:
      g1 = gfapy.Gfa.from_file("tests/testdata/links_distri.l2.{}".format(sfx))
      g2 = gfapy.Gfa.from_file("tests/testdata/links_distri.l2.m3.{}".format(sfx))
      self.assertNotEqual(set(g2.segment_names),set(g1.segment_names))
      self.assertNotEqual(set([str(x) for x in g2.dovetails]),
                          set([str(x) for x in g1.dovetails]))
      g1.multiply("1", 3, extended=True)
      self.assertEqual(set(g2.segment_names),set(g1.segment_names))
      self.assertEqual(set([str(x) for x in g2.dovetails]),
                       set([str(x) for x in g1.dovetails]))

  def test_no_links_distribution_l2_m3(self):
    for sfx in ["gfa", "gfa2"]:
      g1 = gfapy.Gfa.from_file("tests/testdata/links_distri.l2.{}".format(sfx))
      g2 = gfapy.Gfa.from_file("tests/testdata/links_distri.l2.m3.no_ld.{}".format(sfx))
      self.assertNotEqual(set(g2.segment_names),set(g1.segment_names))
      self.assertNotEqual(set([str(x) for x in g2.dovetails]),
                          set([str(x) for x in g1.dovetails]))
      g1.multiply("1", 3, extended=True, distribute="off")
      self.assertEqual(set(g2.segment_names),set(g1.segment_names))
      self.assertEqual(set([str(x) for x in g2.dovetails]),
                       set([str(x) for x in g1.dovetails]))

  def test_links_distribution_l3_m2(self):
    for sfx in ["gfa", "gfa2"]:
      g1 = gfapy.Gfa.from_file("tests/testdata/links_distri.l3.{}".format(sfx))
      g2 = gfapy.Gfa.from_file("tests/testdata/links_distri.l3.m2.{}".format(sfx))
      self.assertNotEqual(set(g2.segment_names),set(g1.segment_names))
      self.assertNotEqual(set([str(x) for x in g2.dovetails]),
                          set([str(x) for x in g1.dovetails]))
      g1.multiply("1", 2, extended=True)
      self.assertEqual(set(g2.segment_names),set(g1.segment_names))
      self.assertEqual(set([str(x) for x in g2.dovetails]),
                       set([str(x) for x in g1.dovetails]))

  def test_no_links_distribution_l3_m2(self):
    for sfx in ["gfa", "gfa2"]:
      g1 = gfapy.Gfa.from_file("tests/testdata/links_distri.l3.{}".format(sfx))
      g2 = gfapy.Gfa.from_file("tests/testdata/links_distri.l3.m2.no_ld.{}".format(sfx))
      self.assertNotEqual(set(g2.segment_names),set(g1.segment_names))
      self.assertNotEqual(set([str(x) for x in g2.dovetails]),
                          set([str(x) for x in g1.dovetails]))
      g1.multiply("1", 2, extended=True, distribute="off")
      self.assertEqual(set(g2.segment_names),set(g1.segment_names))
      self.assertEqual(set([str(x) for x in g2.dovetails]),
                       set([str(x) for x in g1.dovetails]))

  def test_muliply_without_rgfatools(self):
    for sfx in ["gfa", "gfa2"]:
      g1 = gfapy.Gfa.from_file("tests/testdata/links_distri.l3.{}".format(sfx))
      g2 = gfapy.Gfa.from_file("tests/testdata/links_distri.l3.m2.no_ld.{}".format(sfx))
      self.assertNotEqual(set(g2.segment_names),set(g1.segment_names))
      self.assertNotEqual(set([str(x) for x in g2.dovetails]),
                          set([str(x) for x in g1.dovetails]))
      g1.multiply("1", 2)
      self.assertEqual(set(g2.segment_names),set(g1.segment_names))
      self.assertEqual(set([str(x) for x in g2.dovetails]),
                       set([str(x) for x in g1.dovetails]))

  def test_distribution_policy_equal_with_equal(self):
    for sfx in ["gfa", "gfa2"]:
      g1 = gfapy.Gfa.from_file("tests/testdata/links_distri.l2.{}".format(sfx))
      g2 = gfapy.Gfa.from_file("tests/testdata/links_distri.l2.m2.{}".format(sfx))
      self.assertNotEqual(set(g2.segment_names),set(g1.segment_names))
      self.assertNotEqual(set([str(x) for x in g2.dovetails]),
                          set([str(x) for x in g1.dovetails]))
      g1.multiply("1", 2, extended=True, distribute="equal")
      self.assertEqual(set(g2.segment_names),set(g1.segment_names))
      self.assertEqual(set([str(x) for x in g2.dovetails]),
                       set([str(x) for x in g1.dovetails]))

  def test_distribution_policy_equal_with_not_equal(self):
    for sfx in ["gfa", "gfa2"]:
      g1 = gfapy.Gfa.from_file("tests/testdata/links_distri.l3.{}".format(sfx))
      g2 = gfapy.Gfa.from_file("tests/testdata/links_distri.l3.m2.no_ld.{}".format(sfx))
      self.assertNotEqual(set(g2.segment_names),set(g1.segment_names))
      self.assertNotEqual(set([str(x) for x in g2.dovetails]),
                          set([str(x) for x in g1.dovetails]))
      g1.multiply("1", 2, extended=True, distribute="equal")
      self.assertEqual(set(g2.segment_names),set(g1.segment_names))
      self.assertEqual(set([str(x) for x in g2.dovetails]),
                       set([str(x) for x in g1.dovetails]))

  def test_distribution_policy_L(self):
    for sfx in ["gfa", "gfa2"]:
      g1 = gfapy.Gfa.from_file("tests/testdata/links_distri.l2.{}".format(sfx))
      g2 = gfapy.Gfa.from_file("tests/testdata/links_distri.l2.m2.no_ld.{}".format(sfx))
      self.assertNotEqual(set(g2.segment_names),set(g1.segment_names))
      self.assertNotEqual(set([str(x) for x in g2.dovetails]),
                          set([str(x) for x in g1.dovetails]))
      g1.multiply("1", 2, extended=True, distribute="L")
      self.assertEqual(set(g2.segment_names),set(g1.segment_names))
      self.assertEqual(set([str(x) for x in g2.dovetails]),
                       set([str(x) for x in g1.dovetails]))

  def test_distribution_policy_R(self):
    for sfx in ["gfa", "gfa2"]:
      g1 = gfapy.Gfa.from_file("tests/testdata/links_distri.l2.{}".format(sfx))
      g2 = gfapy.Gfa.from_file("tests/testdata/links_distri.l2.m2.{}".format(sfx))
      self.assertNotEqual(set(g2.segment_names),set(g1.segment_names))
      self.assertNotEqual(set([str(x) for x in g2.dovetails]),
                          set([str(x) for x in g1.dovetails]))
      g1.multiply("1", 2, extended=True, distribute="R")
      self.assertEqual(set(g2.segment_names),set(g1.segment_names))
      self.assertEqual(set([str(x) for x in g2.dovetails]),
                       set([str(x) for x in g1.dovetails]))




import unittest
import gfapy

class TestUnitUnknown(unittest.TestCase):

  u = gfapy.line.Unknown([None, "a"])

  def test_new(self):
    assert(isinstance(TestUnitUnknown.u, gfapy.line.Unknown))

  def test_str(self):
    self.assertEqual("?record_type?\ta\tco:Z:line_created_by_gfapy",
        str(TestUnitUnknown.u))

  def test_tags(self):
    with self.assertRaises(AttributeError):
      TestUnitUnknown.u.xx
    self.assertEqual(None, TestUnitUnknown.u.get("xx"))
    with self.assertRaises(gfapy.RuntimeError):
      TestUnitUnknown.u.xx = 1
    self.assertRaises(gfapy.RuntimeError,
        TestUnitUnknown.u.set,"xx",1)

  def test_virtual(self):
    assert(TestUnitUnknown.u.virtual)



import gfapy

class References:
  def _initialize_references(self):
    for snum in [1,2]:
      sid = "sid{}".format(snum)
      orient = self.get(sid).orient
      linesymbol = self.get(sid).line
      s = self._gfa.segment(linesymbol)
      if s is None:
        if self._gfa._segments_first_order:
          raise gfapy.NotFoundError()
        s = gfapy.line.segment.GFA2({"sid" : linesymbol,
                                     "slen" : 1,
                                     "sequence" : "*"},
                                    version = "gfa2",
                                    virtual = True)
        s.connect(self._gfa)
      self._set_existing_field(sid, gfapy.OrientedLine(s,orient),
          set_reference = True)
      s._add_reference(self, self._refkey_for_s(snum))

  def _refkey_for_s(self, snum):
    a = [self.sid1.orient, self.sid2.orient]
    if a == ["+", "+"]:
      return "gaps_R" if (snum == 1) else "gaps_L"
    elif a == ["+", "-"]:
      return "gaps_R"
    elif a == ["-", "+"]:
      return "gaps_L"
    elif a == ["-", "-"]:
      return "gaps_L" if (snum == 1) else "gaps_R"
    else:
      raise gfapy.AssertionError("Bug found, please report\n"+
                                 "snum: {}".format(snum))

  def _import_field_references(self, previous):
    for sid in ["sid1", "sid2"]:
      orient = self.get(sid).orient
      linesymbol = self.get(sid).line
      self._set_existing_field(sid,
          gfapy.OrientedLine(self._gfa.segment(linesymbol),orient),
          set_reference = True)



import gfapy
import gfapy.line
from .collections import Collections
from .headers import Headers
from .collections import Collections
from .creators import Creators
from .destructors import Destructors
from .finders import Finders

class Lines(Collections, Creators, Destructors, Finders, Headers):

  GFA1Specific = [
                   gfapy.line.edge.Link,
                   gfapy.line.edge.Containment,
                   gfapy.line.group.Path,
                   gfapy.line.segment.GFA1
                 ]

  GFA2Specific = [
                   gfapy.line.CustomRecord,
                   gfapy.line.Fragment,
                   gfapy.line.Gap,
                   gfapy.line.edge.GFA2,
                   gfapy.line.segment.GFA2,
                   gfapy.line.group.Unordered,
                   gfapy.line.group.Ordered,
                   gfapy.line.Unknown
                 ]

  def _api_private_check_gfa_line(self, gfa_line, callermeth):
    if not isinstance(gfa_line, gfapy.Line):
      raise gfapy.TypeError("Note: {} is API private, ".format(callermeth)+
          "do not call it directly\n"+
          "Error: line class is {} and not gfapy.Line")
    elif not gfa_line._gfa is self:
      raise gfapy.RuntimeError("Note: {} is API private, ".format(callermeth)+
          "do not call it directly\n"+
          "Error: line.gfa is not the expected instance of gfapy.Gfa\n"+
          repr(gfa_line.gfa)+" != "+repr(self))



import gfapy

class Equivalence:

  def __hash__(self):
    """
    Computes an hash for including the link in a dict,
    so that the hash of a link and its complement is the same.
    Thereby, tags are not considered.
    """
    hash(str(self.from_end)) + \
    hash(str(to_end)) + \
    hash(str(overlap)) + \
    hash(str(overlap.complement()))

  def is_eql(self, other):
    """
    Compares two links and determine their equivalence.
    Thereby, tags are not considered.

    .. note::
      Inverting the strand of both links and reversing
      the CIGAR operations (order/type), one obtains an
      equivalent complement link.

    Parameters
    ----------
    other : gfapy.line.edge.Link
      A link.

    Returns
    -------
    bool
      Are self and other equivalent?

    See Also
    --------
    ==
    is_same
    is_complement
    """
    return (self.is_same(other) or self.is_complement(other))

  def are_tags_eql(self, other):
    """
    Compares the tags of two links.

    .. note::
      This method shall be overridden if custom tags
      are defined, which have a complementation operation which determines
      their value in the equivalent but complement link.

    Parameters
    ----------
    other : gfapy.line.edge.Link
      A link.

    Returns
    -------
    bool
      Are self and other equivalent?

    See Also
    --------
    __eq__
    """
    return (sorted(self.tagnames) == sorted(other.tagnames)) and \
            all(self.get(fn) == other.get(fn) for fn in self.tagnames)

#  def __eq__(self, other):
#    """
#    Compares two links and determine their equivalence.
#    Tags must have the same content.
#
#    .. note:: Inverting the strand of both links and reversing
#    the CIGAR operations (order/type), one obtains an equivalent
#    link.
#
#    Parameters
#    ----------
#    other : gfapy.line.edge.Link
#      A link.
#
#    Returns
#    -------
#    bool
#      Are self and other equivalent?
#
#    See Also
#    --------
#    is_eql
#    are_tags_eql
#    """
#    return self.is_eql(other) and self.are_tags_eql(other)

  def is_same(self, other):
    """
    Compares two links and determine their equivalence.
    Thereby, tags are not considered.

    Parameters
    ----------
    other : gfapy.line.edge.Link
      A link.

    Returns
    -------
    bool
      Are self and other equivalent?

    See Also
    --------
    is_eql
    is_complement
    __eq__
    """
    return (self.from_end == other.from_end and
            self.to_end == other.to_end and
            self.overlap == other.overlap)

  def is_complement(self, other):
    """
    Compares the link to the complement of another link
    and determine their equivalence.
    Thereby, tags are not considered.

    Parameters
    ----------
    other : gfapy.line.edge.Link
      The other link.

    Returns
    -------
    bool
      Are self and the complement of other equivalent?

    See Also
    --------
    is_eql
    is_same
    __eq__
    """
    return (self.from_end == other.to_end and
            self.to_end == other.from_end and
            self.overlap == other.overlap.complement())

  def is_compatible(self, other_oriented_from, other_oriented_to,
                    other_overlap = None, allow_complement = True):
    """
    Compares a link and optionally the complement link,
    with two oriented_segments and optionally an overlap.

    Parameters
    ----------
    other_oriented_from : gfapy.OrientedLine
    other_oriented_to : gfapy.OrientedLine
    allow_complement : bool
      Shall the complement link also be considered?
    other_overlap : gfapy.Alignment.CIGAR
      Compared only if not empty.

    Returns
    -------
    bool
      Does the link or, if **allow_complement**, the complement link go from
      the first, oriented segment to the second with an overlap equal to the
      provided one (if not empty)?
    """
    other_overlap = gfapy.Alignment(other_overlap, version = "gfa1",
        valid = True)
    if self.is_compatible_direct(other_oriented_from, other_oriented_to,
        other_overlap):
      return True
    elif allow_complement:
      return self.is_compatible_complement(other_oriented_from,
                                           other_oriented_to,
                                           other_overlap)
    else:
      return False

  def is_compatible_direct(self, other_oriented_from, other_oriented_to,
                           other_overlap = None):
    """
    Compares a link with two oriented segments and optionally an overlap.

    Parameters
    ----------
    other_oriented_from : gfapy.OrientedLine
    other_oriented_to : gfapy.OrientedLine
    other_overlap : gfapy.Alignment.CIGAR
      Compared only if not empty.

    Returns
    -------
    bool
      Does the link go from the first oriented segment to the second
      with an overlap equal to the provided one (if not empty)?
    """
    return ((self.oriented_from == other_oriented_from and
             self.oriented_to == other_oriented_to) and
           (not self.overlap or not other_overlap or
             (self.overlap == other_overlap)))

  def is_compatible_complement(self, other_oriented_from, other_oriented_to,
                               other_overlap = None):
    """
    Compares the complement link with two oriented segments and optionally an
    overlap.

    Parameters
    ----------
    other_oriented_from : gfapy.OrientedLine
    other_oriented_to : gfapy.OrientedLine
    other_overlap : gfapy.Alignment.CIGAR
      Compared only if not empty.

    Returns
    -------
    bool
      Does the complement link go from the first oriented segment
      to the second with an overlap equal to the provided one (if not empty)?
    """
    return ((self.oriented_to == other_oriented_from.inverted() and
            (self.oriented_from == other_oriented_to.inverted()) and
            (not self.overlap or not other_overlap or
            (self.overlap == other_overlap.complement()))))

  def _complement_ends(self, other):
    return (self.from_end == other.to_end and self.to_end == other.from_end)



import gfapy
import unittest

class TestApiGfa1Lines(unittest.TestCase):

  def test_C(self):
    fields=["C","1","+","2","-","12","12M","MQ:i:1232","NM:i:3","ab:Z:abcd"]
    s="\t".join(fields)
    gfapy.Line(s) # nothing raised
    self.assertEqual(gfapy.line.edge.Containment, gfapy.Line(s).__class__)
    self.assertEqual(fields[0], gfapy.Line(s).record_type)
    self.assertEqual(fields[1], gfapy.Line(s).from_segment)
    self.assertEqual(fields[2], gfapy.Line(s).from_orient)
    self.assertEqual(fields[3], gfapy.Line(s).to_segment)
    self.assertEqual(fields[4], gfapy.Line(s).to_orient)
    self.assertEqual(12, gfapy.Line(s).pos)
    self.assertEqual([gfapy.CIGAR.Operation(12,"M")], gfapy.Line(s).overlap)
    self.assertEqual(1232, gfapy.Line(s).MQ)
    self.assertEqual(3, gfapy.Line(s).NM)
    self.assertEqual("abcd", gfapy.Line(s).ab)
    with self.assertRaises(gfapy.FormatError): (str+gfapy.Line("\tH1"))
    with self.assertRaises(gfapy.FormatError): gfapy.Line("C\tH")
    with self.assertRaises(gfapy.FormatError):
      f=fields.copy(); f[2]="x"; gfapy.Line("\t".join(f),vlevel=1)
    with self.assertRaises(gfapy.FormatError):
      f=fields.copy(); f[4]="x"; gfapy.Line("\t".join(f),vlevel=1)
    with self.assertRaises(gfapy.FormatError):
      f=fields.copy(); f[5]="x"; gfapy.Line("\t".join(f),vlevel=1)
    with self.assertRaises(gfapy.FormatError):
      f=fields.copy(); f[6]="x"; gfapy.Line("\t".join(f),vlevel=1)
    with self.assertRaises(gfapy.TypeError):
      f=fields.copy(); f[7]="MQ:Z:1232"; gfapy.Line("\t".join(f),vlevel=1)
    with self.assertRaises(gfapy.TypeError):
      f=fields.copy(); f[8]="NM:Z:1232"; gfapy.Line("\t".join(f),vlevel=1)

  def test_L(self):
    fields=["L","1","+","2","-","12M","RC:i:1232","NM:i:3","ab:Z:abcd",
            "FC:i:2321","KC:i:1212","MQ:i:40"]
    s="\t".join(fields)
    gfapy.Line(s) # nothing raised
    self.assertEqual(gfapy.line.edge.Link, gfapy.Line(s).__class__)
    self.assertEqual(fields[0], gfapy.Line(s).record_type)
    self.assertEqual(fields[1], gfapy.Line(s).from_segment)
    self.assertEqual(fields[2], gfapy.Line(s).from_orient)
    self.assertEqual(fields[3], gfapy.Line(s).to_segment)
    self.assertEqual(fields[4], gfapy.Line(s).to_orient)
    self.assertEqual([gfapy.CIGAR.Operation(12,"M")],
                 gfapy.Line(s).overlap)
    self.assertEqual(1232, gfapy.Line(s).RC)
    self.assertEqual(3, gfapy.Line(s).NM)
    self.assertEqual(2321, gfapy.Line(s).FC)
    self.assertEqual(1212, gfapy.Line(s).KC)
    self.assertEqual(40, gfapy.Line(s).MQ)
    self.assertEqual("abcd", gfapy.Line(s).ab)
    with self.assertRaises(gfapy.FormatError): (str+gfapy.Line("\tH1"))
    with self.assertRaises(gfapy.FormatError): gfapy.Line("L\tH")
    with self.assertRaises(gfapy.FormatError):
      f=fields.copy(); f[2]="x"; gfapy.Line("\t".join(f),vlevel=1)
    with self.assertRaises(gfapy.FormatError):
      f=fields.copy(); f[4]="x"; gfapy.Line("\t".join(f),vlevel=1)
    with self.assertRaises(gfapy.FormatError):
      f=fields.copy(); f[5]="x"; gfapy.Line("\t".join(f),vlevel=1)
    with self.assertRaises(gfapy.TypeError):
      f=fields.copy(); f[6]="RC:Z:1232"; gfapy.Line("\t".join(f),vlevel=1)
    with self.assertRaises(gfapy.TypeError):
      f=fields.copy(); f[7]="NM:Z:1232"; gfapy.Line("\t".join(f),vlevel=1)

  def test_L_coords(self):
    g = gfapy.Gfa(version="gfa1")
    g.append("S\t1\t*\tLN:i:100")
    g.append("L\t1\t+\t2\t-\t1M2D10M1I")
    self.assertEqual(["87","100$"], [str(s) for s in g.dovetails[0].from_coords])
    with self.assertRaises(gfapy.ValueError): g.dovetails[0].to_coords
    g.append("S\t2\t*\tLN:i:100")
    self.assertEqual(["88","100$"], [str(s) for s in g.dovetails[0].to_coords])
    g.append("L\t3\t-\t4\t+\t10M2P3D1M")
    self.assertEqual(["0","14"], [str(s) for s in g.dovetails[1].from_coords])
    self.assertEqual(["0","11"], [str(s) for s in g.dovetails[1].to_coords])

  def test_L_other(self):
    l = gfapy.Line("L\t1\t+\t2\t-\t*")
    self.assertEqual("2", l.other("1"))
    self.assertEqual("1", l.other("2"))
    with self.assertRaises(gfapy.NotFoundError): l.other("0")

  def test_L_circular(self):
    l = gfapy.Line("L\t1\t+\t2\t-\t*")
    self.assertEqual(False, l.is_circular())
    l = gfapy.Line("L\t1\t+\t1\t-\t*")
    self.assertEqual(True, l.is_circular())

  def test_S(self):
    fields=["S","1","ACGTCACANNN","RC:i:1232","LN:i:11","ab:Z:abcd",
            "FC:i:2321","KC:i:1212"]
    s="\t".join(fields)
    gfapy.Line(s) # nothing raised
    self.assertEqual(gfapy.line.segment.GFA1, gfapy.Line(s).__class__)
    self.assertEqual(fields[0], gfapy.Line(s).record_type)
    self.assertEqual(fields[1], gfapy.Line(s).name)
    self.assertEqual(fields[2], gfapy.Line(s).sequence)
    self.assertEqual(1232, gfapy.Line(s).RC)
    self.assertEqual(11, gfapy.Line(s).LN)
    self.assertEqual(2321, gfapy.Line(s).FC)
    self.assertEqual(1212, gfapy.Line(s).KC)
    self.assertEqual("abcd", gfapy.Line(s).ab)
    with self.assertRaises(gfapy.FormatError): (str+gfapy.Line("\tH1"))
    with self.assertRaises(gfapy.FormatError): gfapy.Line("S\tH")
    with self.assertRaises(gfapy.FormatError):
      f=fields.copy(); f[2]="!@#?"; gfapy.Line("\t".join(f),vlevel=1)
    with self.assertRaises(gfapy.TypeError):
      f=fields.copy(); f[3]="RC:Z:1232"; gfapy.Line("\t".join(f),version="gfa1")
    f=["S","2","ACGTCACANNN","LN:i:3"]
    with self.assertRaises(gfapy.InconsistencyError):
      gfapy.Line("\t".join(f),vlevel=1, version="gfa1")
    f=["S","2","ACGTCACANNN","LN:i:11"]
    gfapy.Line("\t".join(f)) # nothing raised
    f=["S","2","*","LN:i:3"]
    gfapy.Line("\t".join(f)) # nothing raised

  def test_forbidden_segment_names(self):
    gfapy.Line("S\tA+B\t*") # nothing raised
    gfapy.Line("S\tA-B\t*") # nothing raised
    gfapy.Line("S\tA,B\t*") # nothing raised
    with self.assertRaises(gfapy.FormatError):
      gfapy.Line("S\tA+,B\t*",vlevel=1)
    with self.assertRaises(gfapy.FormatError):
      gfapy.Line("S\tA-,B\t*",vlevel=1)

  def test_coverage(self):
    l = gfapy.Line("S\t0\t*\tRC:i:600\tLN:i:100")
    self.assertEqual(6, l.coverage())
    self.assertEqual(6, l.try_get_coverage())
    l = gfapy.Line("S\t0\t*\tRC:i:600")
    self.assertEqual(None, l.coverage())
    with self.assertRaises(gfapy.NotFoundError): l.try_get_coverage()
    l = gfapy.Line("S\t0\t*\tLN:i:100")
    self.assertEqual(None, l.coverage())
    with self.assertRaises(gfapy.NotFoundError): l.try_get_coverage()
    l = gfapy.Line("S\t0\t*\tFC:i:600\tLN:i:100")
    self.assertEqual(None, l.coverage())
    with self.assertRaises(gfapy.NotFoundError): l.try_get_coverage()
    self.assertEqual(6, l.coverage(count_tag="FC"))
    self.assertEqual(6, l.try_get_coverage(count_tag="FC"))

  def test_P(self):
    fields=["P","4","1+,2-,3+","9M2I3D1M,12M","ab:Z:abcd"]
    s="\t".join(fields)
    gfapy.Line(s) # nothing raised
    self.assertEqual(gfapy.line.group.Path, gfapy.Line(s).__class__)
    self.assertEqual(fields[0], gfapy.Line(s).record_type)
    self.assertEqual(fields[1], gfapy.Line(s).path_name)
    self.assertEqual([gfapy.OrientedLine("1","+"),gfapy.OrientedLine("2","-"),
        gfapy.OrientedLine("3","+")],
                 gfapy.Line(s).segment_names)
    self.assertEqual([[gfapy.CIGAR.Operation(9,"M"),
                   gfapy.CIGAR.Operation(2,"I"),
                   gfapy.CIGAR.Operation(3,"D"),
                   gfapy.CIGAR.Operation(1,"M")],
                  [gfapy.CIGAR.Operation(12,"M")]],
                 gfapy.Line(s).overlaps)
    self.assertEqual("abcd", gfapy.Line(s).ab)
    with self.assertRaises(gfapy.FormatError): (str+gfapy.Line("\tH1"))
    with self.assertRaises(gfapy.FormatError): gfapy.Line("P\tH")
    with self.assertRaises(gfapy.FormatError):
      f=fields.copy(); f[2]="1,2,3"; gfapy.Line("\t".join(f),vlevel=1)
    with self.assertRaises(gfapy.InconsistencyError):
      f=fields.copy(); f[2]="1+,2+"; f[3]="9M,12M,3M"
      gfapy.Line("\t".join(f),vlevel=1)
      f=fields.copy(); f[3]="*,*";
      gfapy.Line("\t".join(f),vlevel=1)
      f=fields.copy(); f[3]="9M2I3D1M,12M,12M";
      gfapy.Line("\t".join(f),vlevel=3)
      f=fields.copy(); f[3]="*";
      gfapy.Line("\t".join(f),vlevel=1)
    with self.assertRaises(gfapy.FormatError):
      f=fields.copy(); f[3]="12,12"; gfapy.Line("\t".join(f),vlevel=1)
    with self.assertRaises(gfapy.FormatError):
      f=fields.copy(); f[3]="12M|12M"; gfapy.Line("\t".join(f),vlevel=1)




import gfapy
import unittest

class TestApiVersion(unittest.TestCase):

  def test_header_conversion(self):
    gfa1str = "H\tVN:Z:1.0"
    gfa2str = "H\tVN:Z:2.0"
    self.assertEqual(gfa1str, str(gfapy.Line(gfa2str).to_gfa1()))
    self.assertEqual(gfa1str, str(gfapy.Line(gfa2str).to_gfa1()))
    self.assertEqual(gfa2str, str(gfapy.Line(gfa1str).to_gfa2()))
    self.assertEqual(gfa2str, str(gfapy.Line(gfa2str).to_gfa2()))

  def test_comment_conversion(self):
    self.assertEqual("# comment",
                 str(gfapy.Line("# comment",version="gfa1").to_gfa1()))
    self.assertEqual("# comment",
                 str(gfapy.Line("# comment",version="gfa2").to_gfa1()))
    self.assertEqual("# comment",
                 str(gfapy.Line("# comment",version="gfa1").to_gfa2()))
    self.assertEqual("# comment",
                 str(gfapy.Line("# comment",version="gfa2").to_gfa2()))

  def test_segment_conversion(self):
    self.assertEqual("S\tA\tNNNN",
        str(gfapy.Line("S\tA\tNNNN").to_gfa1()))
    self.assertEqual("S\tA\t4\tNNNN",
        str(gfapy.Line("S\tA\tNNNN").to_gfa2()))
    self.assertEqual("S\tA\tNNNN\tLN:i:4",str(gfapy.Line("S\tA\t4\tNNNN").to_gfa1()))
    self.assertEqual("S\tA\t*\tLN:i:4",str(gfapy.Line("S\tA\t4\t*").to_gfa1()))
    self.assertEqual("S\tA\t4\tNNNN",str(gfapy.Line("S\tA\t4\tNNNN").to_gfa2()))
    # wrong sequence alphabet for GFA2->GFA1
    self.assertEqual("S\tA\t4\t[[]]",str(gfapy.Line("S\tA\t4\t[[]]").to_gfa2()))
    self.assertRaises(gfapy.RuntimeError,gfapy.Line("S\tA\t4\t[[]]").to_gfa1)
    # wrong identifier for GFA2->GFA1
    self.assertEqual("S\tA+,\t3\tNNN", str(gfapy.Line("S\tA+,\t3\tNNN").to_gfa2()))
    self.assertRaises(gfapy.RuntimeError,gfapy.Line("S\tA+,\t3\tNNN").to_gfa1)
    # sequence not available but LN for GFA1->GFA2
    self.assertEqual("S\tA\t4\t*",str(gfapy.Line("S\tA\t*\tLN:i:4").to_gfa2()))
    # both sequence and LN not available for GFA1->GFA2
    self.assertRaises(gfapy.RuntimeError,gfapy.Line("S\tA\t*").to_gfa2)

  def test_link_conversion(self):
    gfa1str = "L\tA\t+\tB\t-\t100M"
    gfa1str_noov = "L\tA\t+\tB\t+\t*"
    gfa2str = "E\t1\tA+\tB-\t100\t200$\t100\t200$\t100M"
    # not connected
    self.assertRaises(gfapy.RuntimeError,gfapy.Line(gfa1str).to_gfa2)
    # connected
    g = gfapy.Gfa()
    g.add_line("S\tA\t*\tLN:i:200")
    g.add_line("S\tB\t*\tLN:i:200")
    gfa1line = gfapy.Line(gfa1str)
    g.add_line(gfa1line)
    gfa1line_noov = gfapy.Line(gfa1str_noov)
    g.add_line(gfa1line_noov)
    self.assertEqual(gfa2str,str(gfa1line.to_gfa2()))
    self.assertEqual(gfa1str+"\tID:Z:1",str(gfa1line.to_gfa1()))
    # placeholder overlap
    self.assertRaises(gfapy.ValueError,gfa1line_noov.to_gfa2)
    # TODO check if the alignment is compatible with the segment length

  def test_containment_conversion(self):
    gfa1str = "C\tA\t+\tB\t-\t20\t100M"
    gfa1str_noov = "C\tA\t+\tB\t+\t20\t*"
    gfa2str = "E\t1\tA+\tB-\t20\t120\t0\t100$\t100M"
    # not connected
    self.assertRaises(gfapy.RuntimeError,gfapy.Line(gfa1str).to_gfa2)
    # connected
    g = gfapy.Gfa()
    g.add_line("S\tA\t*\tLN:i:200")
    g.add_line("S\tB\t*\tLN:i:100")
    gfa1line = gfapy.Line(gfa1str)
    g.add_line(gfa1line)
    gfa1line_noov = gfapy.Line(gfa1str_noov)
    g.add_line(gfa1line_noov)
    self.assertEqual(gfa2str,str(gfa1line.to_gfa2()))
    self.assertEqual(gfa1str+"\tID:Z:1",str(gfa1line.to_gfa1()))
    # placeholder overlap
    self.assertRaises(gfapy.ValueError,gfa1line_noov.to_gfa2)
    # TODO check if the alignment is compatible with the segment length

  def test_edge_conversion(self):
    dovetail         = "E\t*\tA+\tB-\t100\t200$\t100\t200$\t100M"
    dovetail_gfa1    = "L\tA\t+\tB\t-\t100M"
    containment      = "E\t*\tA+\tB-\t20\t120\t0\t100$\t100M"
    containment_gfa1 = "C\tA\t+\tB\t-\t20\t100M"
    internal         = "E\t*\tA+\tB-\t20\t110\t10\t100$\t90M"
    self.assertEqual(dovetail_gfa1,str( gfapy.Line(dovetail).to_gfa1()))
    self.assertEqual(containment_gfa1,str( gfapy.Line(containment).to_gfa1()))
    self.assertRaises(gfapy.RuntimeError,gfapy.Line(internal).to_gfa1)

  def test_L_to_E(self):
    g = gfapy.Gfa(version="gfa1")
    g.add_line("S\t1\t*\tLN:i:100")
    g.add_line("S\t2\t*\tLN:i:100")
    g.add_line("S\t3\t*\tLN:i:100")
    g.add_line("S\t4\t*\tLN:i:100")
    g.add_line("L\t1\t+\t2\t+\t10M")
    g.add_line("L\t1\t-\t2\t-\t20M")
    g.add_line("L\t3\t-\t4\t+\t30M")
    g.add_line("L\t3\t+\t4\t-\t40M")
    dovetails_gfa1 = g.dovetails
    dovetails_gfa2 = {dovetails_gfa1[0].to_gfa2_s(),
                      dovetails_gfa1[1].to_gfa2_s(),
                      dovetails_gfa1[2].to_gfa2_s(),
                      dovetails_gfa1[3].to_gfa2_s()}
    expected_dovetails_gfa2 = {
        "E	5	1+	2+	90	100$	0	10	10M",
        "E	6	1-	2-	0	20	80	100$	20M",
        "E	7	3-	4+	0	30	0	30	30M",
        "E	8	3+	4-	60	100$	60	100$	40M"}
    try:
      self.assertEqual(expected_dovetails_gfa2, dovetails_gfa2)
    except:
      # sometimes 7 and 8 are assigned with a different order
      # despite using a fixed hash seed in the tests
      expected_dovetails_gfa2 = {
          "E	5	1+	2+	90	100$	0	10	10M",
          "E	6	1-	2-	0	20	80	100$	20M",
          "E	8	3-	4+	0	30	0	30	30M",
          "E	7	3+	4-	60	100$	60	100$	40M"}
      self.assertEqual(expected_dovetails_gfa2, dovetails_gfa2)
    assert(isinstance(g.dovetails[0].to_gfa1(),gfapy.line.edge.Link))
    assert(isinstance(g.dovetails[0].to_gfa2(),gfapy.line.edge.GFA2))

  def test_E_to_L(self):
    e1 = gfapy.Line("E\t*\t1+\t2+\t90\t100$\t0\t10\t10M")
    l1 = "L\t1\t+\t2\t+\t10M"
    self.assertEqual(l1, e1.to_gfa1_s())
    e2 = gfapy.Line("E\t*\t1+\t2+\t0\t20\t80\t100$\t20M")
    l2 = "L\t2\t+\t1\t+\t20M"
    self.assertEqual(l2, e2.to_gfa1_s())
    e3 = gfapy.Line("E\t*\t3-\t4+\t0\t30\t0\t30\t30M")
    l3 = "L\t3\t-\t4\t+\t30M"
    self.assertEqual(l3, e3.to_gfa1_s())
    e4 = gfapy.Line("E\t*\t3+\t4-\t60\t100$\t60\t100$\t40M")
    l4 = "L\t3\t+\t4\t-\t40M"
    self.assertEqual(l4, e4.to_gfa1_s())

  def test_path_conversion(self):
    path_gfa1 = "P\t1\ta+,b-\t100M"
    path_gfa2 = "O\t1\ta+ a_to_b+ b-"
    # gfa1 => gfa2
    l1 = "L\ta\t+\tb\t-\t100M\tID:Z:a_to_b"
    g1 = gfapy.Gfa()
    path_gfa1_line = gfapy.Line(path_gfa1)
    g1.add_line(path_gfa1_line)
    g1.add_line(l1)
    g1.process_line_queue()
    # not connected
    self.assertRaises(gfapy.RuntimeError,
        gfapy.Line(path_gfa1).to_gfa2)
    # connected
    self.assertEqual(path_gfa1,str(path_gfa1_line.to_gfa1()))
    self.assertEqual(path_gfa2,str(path_gfa1_line.to_gfa2()))
    # gfa2 => gfa1
    e = "E\ta_to_b\ta+\tb-\t100\t200$\t100\t200$\t100M"
    sA = "S\ta\t200\t*"
    sB = "S\tb\t200\t*"
    g2 = gfapy.Gfa()
    path_gfa2_line = gfapy.Line(path_gfa2)
    g2.add_line(path_gfa2_line)
    g2.add_line(e)
    g2.add_line(sA)
    g2.add_line(sB)
    # not connected
    self.assertRaises(gfapy.RuntimeError,
        gfapy.Line(path_gfa2).to_gfa1)
    # connected
    self.assertEqual(path_gfa1,str( path_gfa2_line.to_gfa1()))
    self.assertEqual(path_gfa2,str( path_gfa2_line.to_gfa2()))

  def test_gap_conversion(self):
    s = "G\t*\tA-\tB+\t100\t*"
    self.assertEqual(s, str(gfapy.Line(s).to_gfa2()))
    self.assertRaises(gfapy.VersionError,gfapy.Line(s).to_gfa1)

  def test_fragment_conversion(self):
    s = "F\tA\tread1-\t0\t100\t0\t100\t*"
    self.assertEqual(s,str( gfapy.Line(s).to_gfa2()))
    self.assertRaises(gfapy.VersionError,gfapy.Line(s).to_gfa1)

  def test_set_conversion(self):
    s = "U\t1\tA B C"
    self.assertEqual(s,str( gfapy.Line(s).to_gfa2()))
    self.assertRaises(gfapy.VersionError,gfapy.Line(s).to_gfa1)

  def test_custom_record_conversion(self):
    s = "X\tx1\tA\tC"
    self.assertEqual(s,str( gfapy.Line(s).to_gfa2()))
    self.assertRaises(gfapy.VersionError,gfapy.Line(s).to_gfa1)

  def test_unknown_record_conversion(self):
    record = gfapy.line.Unknown([None, "A"])
    self.assertEqual(record, record.to_gfa2())
    self.assertRaises(gfapy.VersionError,record.to_gfa1)

  def test_gfa_conversion(self):
    gfa1_str ='''# comment
H\tVN:Z:1.0
S\tB\t*\tLN:i:200
S\tC\t*\tLN:i:100
S\tA\t*\tLN:i:200
L\tA\t+\tB\t-\t100M\tID:Z:a_to_b
C\tA\t+\tC\t-\t20\t100M\tID:Z:2
P\t1\tA+,B-\t100M'''
    gfa2_str ='''# comment
H\tVN:Z:2.0
S\tB\t200\t*
S\tC\t100\t*
S\tA\t200\t*
E\ta_to_b\tA+\tB-\t100\t200$\t100\t200$\t100M
E\t2\tA+\tC-\t20\t120\t0\t100$\t100M
O\t1\tA+ a_to_b+ B-'''
    self.assertEqual(gfa2_str, str(gfapy.Gfa(gfa1_str).to_gfa2()))
    self.assertEqual(gfa1_str, str(gfapy.Gfa(gfa2_str).to_gfa1()))



import unittest
import gfapy

class TestApiHeader(unittest.TestCase):

  def test_gfa_header(self):
    g = gfapy.Gfa()
    assert(isinstance(g.header, gfapy.line.Header))
    self.assertEqual([], g.header.tagnames)
    g.add_line("H\txx:i:1")
    self.assertEqual(["xx"], g.header.tagnames)

  def test_gfa_header_line_connect(self):
    g = gfapy.Gfa()
    line = gfapy.Line("H\txx:i:1")
    self.assertRaises(gfapy.RuntimeError, line.connect, g)
    g.add_line(line) # nothing raised

  def test_header_version_editing(self):
    standalone = gfapy.Line("H\txx:i:1\tVN:Z:1.0")
    standalone.VN = "2.0" # nothing raised
    g = gfapy.Gfa()
    g.add_line("H\txx:i:1\tVN:Z:1.0")
    g.header.xx = 2 # nothing raised
    with self.assertRaises(gfapy.RuntimeError):
      g.header.VN = "2.0"

  def test_error_inconsistent_definitions(self):
    g = gfapy.Gfa()
    g.add_line("H\txx:i:1")
    g.add_line("H\txx:i:2") # nothing raised
    g.add_line("H\tTS:i:120")
    g.add_line("H\tTS:i:120") # nothing raised
    self.assertRaises(gfapy.InconsistencyError, g.add_line, "H\tTS:i:122")

  def test_gfa_multiple_def_tags(self):
    g = gfapy.Gfa()
    for i in range(4):
      g.add_line("H\txx:i:{}".format(i))
    self.assertEqual(["xx"], g.header.tagnames)
    self.assertEqual([0,1,2,3], g.header.xx)
    self.assertEqual([0,1,2,3], g.header.get("xx"))
    self.assertEqual("i", g.header.get_datatype("xx"))
    g.header.validate_field("xx") # nothing raised
    for i in [0,2,3]:
      g.header.xx.remove(i)
    g.header.xx = [1, 4]
    self.assertRaises(gfapy.TypeError, g.header.validate_field, "xx")
    g.header.xx = gfapy.FieldArray("i", data = g.header.xx)
    g.header.validate_field("xx") # nothing raised
    self.assertEqual([1,4], g.header.get("xx"))
    self.assertEqual("1\t4", g.header.field_to_s("xx"))
    self.assertEqual("xx:i:1\txx:i:4", g.header.field_to_s("xx", tag=True))
    self.assertEqual(sorted(["H\txx:i:1","H\txx:i:4"]),
                     sorted([str(h) for h in g.headers]))
    g.header.add("xx", 12)
    g.header.add("yy", 13)
    self.assertEqual([1,4,12], g.header.xx)
    self.assertEqual(13, g.header.yy)

  def test_gfa_single_def_tags(self):
    g = gfapy.Gfa()
    g.add_line("H\txx:i:1")
    self.assertEqual(["xx"], g.header.tagnames)
    self.assertEqual(1, g.header.xx)
    g.header.set("xx", 12)
    self.assertEqual(12, g.header.xx)
    g.header.delete("xx")
    self.assertEqual(None, g.header.xx)



import gfapy
import unittest

class TestSegmentReferences(unittest.TestCase):

  def test_link_other(self):
    l = gfapy.Line("L\t1\t+\t2\t-\t*")
    self.assertEqual("2", l.other("1"))
    self.assertEqual("1", l.other("2"))
    self.assertRaises(gfapy.NotFoundError, l.other, "0")

  def test_link_circular(self):
    l = gfapy.Line("L\t1\t+\t2\t-\t*")
    self.assertEqual(False, l.is_circular())
    l = gfapy.Line("L\t1\t+\t1\t-\t*")
    self.assertEqual(True, l.is_circular())



import gfapy

class ToGFA1:

  def _to_gfa1_a(self):
    """List of the field content of the line in GFA1.
    """
    at = self._alignment_type
    if at == "I":
      raise gfapy.RuntimeError(
        "Conversion of edge line from GFA2 to GFA1 failed\n"+
        "Edge represents an internal overlap:\n"+
        "Edge line: {}\n".format(str(self)))
    a = [ at ]
    if self._is_sid1_from():
      ol1 = self.get("sid1")
      ol2 = self.get("sid2")
    else:
      ol1 = self.get("sid2")
      ol2 = self.get("sid1")
    a.append(ol1.name)
    a.append(ol1.orient)
    a.append(ol2.name)
    a.append(ol2.orient)
    if self._alignment_type == "C":
      a.append(str(self.pos))
    try:
      self.overlap.validate(version = "gfa1")
    except:
      raise gfapy.RuntimeError(
        "Conversion of edge line from GFA2 to GFA1 failed\n"+
        "Overlap is invalid or not compatible with GFA1\n"+
        "Edge line: {}\n".format(str(self)))
    a.append(str(self.overlap))
    if not gfapy.is_placeholder(self.eid):
      a.append(gfapy.Field._to_gfa_tag(self.eid, "ID", datatype = "Z"))
    for fn in self.tagnames:
      a.append(self.field_to_s(fn, tag = True))
    return a

  @property
  def overlap(self):
    """Value of the GFA1 **overlap** field, if the edge is a link or containment.

    Returns:
      gfapy.Alignment.Placeholder or gfapy.Alignment.CIGAR

    Raises:
      gfapy.error.ValueError: If the edge is internal
    """
    self._check_not_internal("overlap")
    return self.alignment.complement() if self._is_sid1_from() else self.alignment

  @property
  def oriented_from(self):
    return self.sid1 if self._is_sid1_from() else self.sid2

  @property
  def oriented_to(self):
    return self.sid2 if self._is_sid1_from() else self.sid1

  @property
  def from_segment(self):
    """Value of the GFA1 **from_segment** field, if the edge is a link or containment.

    Returns:
      str or gfapy.line.segment.GFA2

    Raises:
      gfapy.error.ValueError: If the edge is internal.
    """
    return self.oriented_from.line

  @from_segment.setter
  def from_segment(self, value):
    """Set the field which will be returned by calling from_segment

    Parameters:
      value (str, gfapy.line.segment.GFA2)
    """
    self.oriented_from.line = value

  @property
  def from_orient(self):
    """Value of the GFA1 **from_orient** field.

    This method can only be applied if the edge is a link or containment.

    Returns:
      str: one of ["+", "-"]

    Raises:
      gfapy.error.ValueError: If the edge is internal.
    """
    return self.oriented_from.orient

  @from_orient.setter
  def from_orient(self, value):
    """Set the orientation of the field which will be returned by calling from

    Parameters:
      value (str): one of ["+", "-"]
    """
    self.oriented_from.orient = value

  @property
  def to_segment(self):
    """Value of the GFA1 ``to_segment`` field.

    This method can only be applied if the edge is a link or containment.

    Returns:
      str or gfapy.line.segment.GFA2

    Raises:
      gfapy.error.ValueError: If the edge is internal.
    """
    return self.oriented_to.line

  @to_segment.setter
  def to_segment(self, value):
    """Set the field which will be returned by calling ``to``.

    Parameters:
      value (str or gfapy.line.segment.GFA2)
    """
    self.oriented_to.line = value

  @property
  def to_orient(self):
    """Value of the GFA1 **to_orient** field.

    This method can only be applied if the edge is a link or containment.

    Returns:
      str : one of ["+", "-"]

    Raises:
      gfapy.error.ValueError: If the edge is internal.
    """
    return self.oriented_to.orient

  @to_orient.setter
  def to_orient(self, value):
    """Set the orientation of the field which will be returned by calling ``to``.

    Parameters:
      value (str): one of ["+", "-"]
    """
    self.oriented_to.orient = value

  @property
  def pos(self):
    """Value of the GFA1 **pos** field, if the edge is a containment.

    Returns:
      int or gfapy.Lastpos

    Raises:
      gfapy.error.ValueError: If the edge is not a containment.
    """
    if self._alignment_type == "I":
      raise gfapy.ValueError("Line: {}\n".format(str(self)) +
                             "Internal alignment, pos is not defined")
    elif self._alignment_type == "L":
      raise gfapy.ValueError("Line: {}\n".format(str(self)) +
                             "Dovetail alignment, pos is not defined")
    elif self._alignment_type == "C":
      if gfapy.isfirstpos(self.beg1):
        return self.beg1 if (gfapy.isfirstpos(self.beg2) and
                             gfapy.islastpos(self.end2)) else self.beg2
      else:
        return self.beg1

  def _check_not_internal(self, fn):
    if self.is_internal():
      raise gfapy.ValueError(
        "Line: {}\n".format(str(self))+
        "Internal alignment, {} is not defined".format(fn))

  @staticmethod
  def _segment_role(begpos, endpos, orient):
    if gfapy.isfirstpos(begpos):
      if gfapy.islastpos(endpos):
        return "contained"
      elif orient == "+":
        return "pfx"
      else:
        return "sfx"
    else:
      if gfapy.islastpos(endpos):
        if orient == "+":
          return "sfx"
        else:
          return "pfx"
      else:
        return "other"

  def _check_GFA1_overlap_compatibility(self):
    pass

  def _is_sid1_from(self):
    sr1 = self._segment_role(self.beg1, self.end1, self.sid1.orient)
    sr2 = self._segment_role(self.beg2, self.end2, self.sid2.orient)
    if sr2 == "contained":
      return True
    elif sr1 == "contained":
      return False
    elif sr1 == "sfx" and sr2 == "pfx":
      return True
    elif sr2 == "sfx" and sr1 == "pfx":
      return False
    else:
      raise gfapy.ValueError(
        "Line: {}\n".format(str(self))+
        "Internal overlap, 'from' is undefined\n"+
        "Roles: segment1 is {} ({},{}), segment2 is {} ({},{})".format(sr1,
          self.beg1, self.end1, sr2, self.beg2, self.end2))




import gfapy

class Validation:

  def validate_positions(self):
    "Checks that positions suffixed by $ are the last position of segments"
    if self.is_connected():
      seg = self.get("sid")
      seq = seg.sequence
      if not gfapy.is_placeholder(seq):
        seqlen = len(seq)
        for sfx in ["beg", "end"]:
          fn = "s_"+sfx
          pos = self.get(fn)
          if gfapy.islastpos(pos):
            if pos != seqlen:
              raise gfapy.InconsistencyError(
                  "Fragment: {}\n".format(str(self))+
                  "Field {}: $ after ".format(str(fn))+
                  "non-last position ({})\n".format(str(pos))+
                  "Segment: {}".format(str(seg)))



import unittest
import gfapy

class TestUnitSymbolInvert(unittest.TestCase):

  def test_invert_orientations(self):
    self.assertEqual("+", gfapy.invert("-"))
    self.assertEqual("-", gfapy.invert("+"))

  def test_invert_segment_ends(self):
    self.assertEqual("L", gfapy.invert("R"))
    self.assertEqual("R", gfapy.invert("L"))

  def test_invert_invalid(self):
    self.assertRaises(gfapy.ValueError, gfapy.invert, "xx")




import gfapy
import unittest

class TestAPIReferencesFGLines(unittest.TestCase):

  def test_fragments_references(self):
    g = gfapy.Gfa()
    f = gfapy.Line("F\ta\tf+\t0\t200\t281\t502$\t*")
    self.assertEqual("a", f.sid)
    self.assertEqual(gfapy.OrientedLine("f","+"), f.external)
    sa = gfapy.Line("S\ta\t100\t*")
    g.append(sa)
    g.append(f)
    self.assertEqual(sa, f.sid)
    f.disconnect()
    self.assertEqual("a", f.sid)
    # disconnection of segment cascades on fragments
    g.append(f)
    assert(f.is_connected())
    self.assertEqual(sa, f.sid)
    sa.disconnect()
    assert(not f.is_connected())
    self.assertEqual("a", f.sid)

  def test_fragments_backreferences(self):
    g = gfapy.Gfa()
    f1 = gfapy.Line("F\ta\tf+\t0\t200\t281\t502$\t*")
    f2 = gfapy.Line("F\ta\tf+\t240\t440$\t0\t210\t*")
    sa = gfapy.Line("S\ta\t100\t*")
    g.append(sa)
    g.append(f1)
    g.append(f2)
    self.assertEqual([f1,f2], sa.fragments)
    # disconnection effects
    f1.disconnect()
    self.assertEqual([f2], sa.fragments)
    sa.disconnect()
    self.assertEqual([], sa.fragments)

  def test_gap_references(self):
    g = gfapy.Gfa()
    gap = gfapy.Line("G\t*\ta+\tb+\t90\t*")
    self.assertEqual(gfapy.OrientedLine("a","+"), gap.sid1)
    self.assertEqual(gfapy.OrientedLine("b","+"), gap.sid2)
    sa = gfapy.Line("S\ta\t100\t*");
    g.append(sa)
    sb = gfapy.Line("S\tb\t100\t*");
    g.append(sb)
    g.append(gap)
    self.assertEqual(sa, gap.sid1.line)
    self.assertEqual(sb, gap.sid2.line)
    gap.disconnect()
    self.assertEqual("a", gap.sid1.line)
    self.assertEqual("b", gap.sid2.line)
    # disconnection of segment cascades on gaps
    g.append(gap)
    assert(gap.is_connected())
    self.assertEqual(sa, gap.sid1.line)
    sa.disconnect()
    assert(not gap.is_connected())
    self.assertEqual("a", gap.sid1.line)

  def test_gaps_backreferences(self):
    g = gfapy.Gfa()
    sa = gfapy.Line("S\ta\t100\t*")
    g.append(sa)
    # gaps
    s = {}
    gap = {}
    for name in ["b", "c", "d", "e", "f", "g", "h", "i"]:
      s[name] = gfapy.Line("S\t{}\t100\t*".format(name))
      g.append(s[name])
    for name in \
      ["a+b+", "a+c-", "a-d+", "a-e-", "f+a+", "g+a-", "h-a+", "i-a-"]:
      gap[name] = gfapy.Line("\t".join(
                                   ["G","*",name[0:2],name[2:4],"200","*"]))
      g.append(gap[name])
    # gaps_[LR]()
    self.assertEqual([gap["a-d+"], gap["a-e-"], gap["f+a+"], gap["h-a+"]],
                 sa.gaps_L)
    self.assertEqual([gap["a+b+"], gap["a+c-"], gap["g+a-"], gap["i-a-"]],
                 sa.gaps_R)
    # gaps()
    self.assertEqual(sa.gaps_L, sa.gaps_of_end("L"))
    self.assertEqual(sa.gaps_R, sa.gaps_of_end("R"))
    self.assertEqual(sa.gaps_L + sa.gaps_R, sa.gaps)
    # disconnection effects
    gap["a-d+"].disconnect()
    self.assertEqual([gap["a-e-"], gap["f+a+"], gap["h-a+"]], sa.gaps_L)
    sa.disconnect()
    self.assertEqual([], sa.gaps_L)
    self.assertEqual([], sa.gaps_R)
    self.assertEqual([], sa.gaps_of_end("L"))
    self.assertEqual([], sa.gaps_of_end("R"))
    self.assertEqual([], sa.gaps)




import gfapy
import re

def unsafe_decode(string):
  return [ gfapy.OrientedLine(str(l[0:-1]), str(l[-1]))
           for l in string.split(",")]

def decode(string):
  validate_encoded(string)
  return unsafe_decode(string)

def validate_encoded(string):
  if not re.match(r"^[!-)+-<>-~][!-~]*[+-](,[!-)+-<>-~][!-~]*[+-])+$", string):
    raise gfapy.FormatError(
      "{} is not a valid list of GFA1 segment names ".format(repr(string))+
      "and orientations\n"+
      "(the segment names must match [!-)+-<>-~][!-~]*;\n"+
      " the orientations must be + or -;\n"+
      " the list must be comma-separated "+
      "NameOrient,NameOrient[,NameOrient...])")

def validate_decoded(iterable):
  for elem in iterable:
    elem = gfapy.OrientedLine(elem)
    elem.validate()
    if not re.match(r"^[!-)+-<>-~][!-~]*$", elem.name):
      raise gfapy.FormatError(
        "#{elem.name} is not a valid GFA1 segment name\n".format(elem.name)+
        "(it does not match [!-)+-<>-~][!-~]*)")

def unsafe_encode(obj):
  if isinstance(obj, str):
    return obj
  elif isinstance(obj, list):
    return ",".join([str(gfapy.OrientedLine(os)) for os in obj])
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: str, list)")

def encode(obj):
  if isinstance(obj, str):
    validate_encoded(obj)
    return obj
  elif isinstance(obj, list):
    validate_decoded(obj)
    return ",".join([str(gfapy.OrientedLine(os)) for os in obj])
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: str, list)")



import gfapy
import re
from collections import OrderedDict

class Taxon(gfapy.Line):
  RECORD_TYPE = "T"
  POSFIELDS = OrderedDict([("tid","identifier_gfa2")])
  TAGS_DATATYPE = {"UL":"Z"}
  NAME_FIELD = "tid"

Taxon.register_extension()

class MetagenomicAssignment(gfapy.Line):
  RECORD_TYPE = "M"
  POSFIELDS = OrderedDict([("mid","optional_identifier_gfa2"),
                           ("tid","identifier_gfa2"),
                           ("sid","identifier_gfa2")])
  TAGS_DATATYPE = {"SC":"i"}
  NAME_FIELD = "mid"

MetagenomicAssignment.register_extension(references=
    [("sid", gfapy.line.segment.GFA2, "metagenomic_assignments"),
     ("tid", Taxon, "metagenomic_assignments")])

class TaxonID:

  def validate_encoded(string):
    if not re.match(r"^taxon:(\d+)$",string) and \
        not re.match(r"^[a-zA-Z0-9_]+$", string):
      raise gfapy.ValueError("Invalid taxon ID: {}".format(string))

  def decode(string):
    TaxonID.validate_encoded(string)
    return string

  def validate_decoded(obj):
    if isinstance(obj,Taxon):
      TaxonID.validate_encoded(obj.name)
    else:
      raise gfapy.TypeError(
        "Invalid type for taxon ID: "+"{}".format(repr(obj)))

  def encode(obj):
    TaxonID.validate_decoded(obj)
    return obj

gfapy.Field.register_datatype("taxon_id", TaxonID)

Taxon.DATATYPE["tid"] = "taxon_id"
MetagenomicAssignment.DATATYPE["tid"] = "taxon_id"



import gfapy
from ..line import Line
from .construction import Construction
from .tags import Tags
from .writer import Writer
from .version_conversion import VersionConversion

class Comment(Writer, Tags, Construction, VersionConversion, Line):
  """
  A comment line of a GFA file

  The content of the comment line, excluding the initial **#** and eventual
  initial spacing characters, is included in the field **content**.

  The initial spacing characters can be read/changed using the **spacer**
  field. The default value is a single space.

  Tags are not supported by comment lines. If the line contains tags,
  these are nor parsed, but included in the **content** field.
  Trying to set or get tag values raises exceptions.
  """

  RECORD_TYPE = "#"
  POSFIELDS = ["content", "spacer"]
  DATATYPE = {
    "content" : "comment",
    "spacer" : "comment",
  }

Comment._apply_definitions()



import gfapy

class References:

  def _initialize_references(self):
    st1 = self._substring_type(self.beg1, self.end1)[0]
    st2 = self._substring_type(self.beg2, self.end2)[0]
    for snum in [1, 2]:
      sid = "sid{}".format(snum)
      orient = self.get(sid).orient
      s = self._gfa.segment(self.get(sid).line)
      if s is None:
        if self._gfa._segments_first_order:
          raise gfapy.NotFoundError()
        s = gfapy.line.segment.GFA2({"sid" : self.get(sid).line,
                                     "slen" : 1,
                                     "sequence" : "*"},
                                    version = "gfa2",
                                    virtual = True)
        s.connect(self._gfa)
      self._set_existing_field(sid, gfapy.OrientedLine(s, orient),
          set_reference = True)
      s._add_reference(self, self._refkey_for_s(snum, st1, st2))

  def _refkey_for_s(self, snum, st1, st2):
    if st1 == "whole":
      if st2 == "whole":
        return "edges_to_contained" if snum == 1 else "edges_to_containers"
      else:
        return "edges_to_containers" if snum == 1 else "edges_to_contained"
    elif st2 == "whole":
      return "edges_to_containers" if snum == 2 else "edges_to_contained"
    elif self.sid1.orient == self.sid2.orient:
      if (st1 == "pfx" and st2 == "sfx"):
        return "dovetails_L" if snum == 1 else "dovetails_R"
      elif (st1 == "sfx" and st2 == "pfx"):
        return "dovetails_R" if snum == 1 else "dovetails_L"
      else:
        return "internals"
    else:
      if (st1 == "pfx" and st2 == "pfx"):
        return "dovetails_L"
      elif (st1 == "sfx" and st2 == "sfx"):
        return "dovetails_R"
      else:
        return "internals"

  def _import_field_references(self, previous):
    for sid in ["sid1", "sid2"]:
      self._set_existing_field(sid,
          gfapy.OrientedLine(self._gfa.segment(self.get(sid).line),
            self.get(sid).orient),
          set_reference = True)

  def _backreference_keys(self, ref, key_in_ref):
    if ref.record_type == "U":
      return ["sets"]
    elif ref.record_type == "O":
      return ["paths"]
    elif ref.record_type == "S":
      return ["sid1", "sid2"]
    else:
      raise gfapy.AssertionError(
        "Bug found, please report\n"+
        "ref: {}\n".format(ref)+
        "key_in_ref: {}".format(key_in_ref))



from ..line import Line
from .connection import Connection
from .multiline import Multiline
from .field_data import FieldData
from .version_conversion import VersionConversion

class Header(VersionConversion, Multiline, Connection, FieldData, Line):
  """
  A header line of a GFA file.

  For examples on how to set the header data, see {GFA.Headers}.

  See Also
  --------
  gfapy.Line
  """

  RECORD_TYPE = "H"
  PREDEFINED_TAGS = ["VN", "TS"]
  DATATYPE = {
    "VN" : "Z",
    "TS" : "i"
  }
  STORAGE_KEY = "merge"

Header._apply_definitions()



import gfapy
import re

class Multiplication:

  def multiply(self, segment, factor, copy_names = None,
               conserve_components = True, distribute = None,
               track_origin = False, origin_tag="or", extended = False):
    """Multiply a segment by a given factor.

    The multiplication operation is implemented as described in
    Gonnella and Kurtz (2016).

    Parameters:
      segment (Line, str): the segment to multiply
      factor (int): the multiplication factor; if 0, the segment is
        deleted; if 1, nothing is done; if > 1, the multiplication
        is performed
      copy_names (list, None): an optional list of strings, the names
        of the copies which will result from the multiplication;
        the length of this list must be equal to factor - 1; if no
        list is specified, the names are computed automatically, adding
        (or incrementing) an integer as suffix to the segment name,
        until enough non-previously used names are found
      conserve_components (bool): if True, the removal of segments
        in the case where factor == 0 is only done if it does not
        split an existing connected component (thereby only dovetail
        overlaps are considered)
      extended : if True, then dovetail distribution and track origin
        are turned on by default
      distribute (str, None) : select an end for which the dovetail
        overlaps are distributed (see Gonnella and Kurtz, 2016); if ``auto``
        (the default if extended is set), an end is selected automatically,
        trying to maximize the number of links which can be deleted; if ``off``
        (the default if extended is not set), no distribution is performed; if
        ``L`` or ``R``, links of the specified end are distributed; if
        ``equal``, an end is selected (if any), for which the number of links
        is equal to the factor (if none, links are not distributed; if both,
        then ``R`` is used)
      track_origin (bool): if True, the name of the original segment (or
        the content of its own origin tag, if any) is stored
        in a tag in the copies (default: False)
      origin_tag (str): the tag where to store the origin information,
        if track_origin is set (default: ``or``)
    """
    if extended:
      if distribute == None:
        distribute = "auto"
      track_origin = True
    if factor < 0:
      raise gfapy.ArgumentError("Mulitiplication factor must be >= 0"+
          " ({} found)".format(factor))
    elif factor == 0:
      if conserve_components and factor == 1 and is_cut_segment(segment):
        return self
      else:
        self.rm(segment)
        return self
    elif factor == 1:
      return self
    else:
      s, sn = self._segment_and_segment_name(segment)
      if track_origin and not s.get(origin_tag):
        s.set(origin_tag, sn)
      self.__divide_segment_and_connection_counts(s, factor)
      if copy_names is None:
        copy_names = self._compute_copy_names(sn, factor)
      for cn in copy_names:
        self.__clone_segment_and_connections(s, cn)
      if distribute:
        self._distribute_links(distribute, sn, copy_names, factor)
      return self

  def _compute_copy_names(self, segment_name, factor):
    assert factor >= 2
    retval = []
    first = 2
    m = re.search(r'(.*)\*(\d+)',segment_name)
    if m:
      segment_name = m.groups()[0]
      i = int(m.groups()[1])
    offset = 0
    for i in range(first,factor+first-1):
      name = "{}*{}".format(segment_name, i+offset)
      while name in self.names:
        offset+=1
        name = "{}*{}".format(segment_name, i+offset)
      retval.append(name)
    return retval

  def __divide_counts(self, gfa_line, factor):
    for count_tag in ["KC", "RC", "FC"]:
      if count_tag in gfa_line.tagnames:
        gfa_line.set(count_tag, gfa_line.get(count_tag) // factor)

  def __divide_segment_and_connection_counts(self, segment, factor):
    self.__divide_counts(segment, factor)
    processed_circulars = set()
    for l in segment.dovetails + segment.containments:
      if l.is_circular():
        if l not in processed_circular:
          self.__divide_counts(l, factor)
          processed_circular.append(l)
      else:
        self.__divide_counts(l, factor)

  def __clone_segment_and_connections(self, segment, clone_name):
    cpy = segment.clone()
    cpy.name = clone_name
    cpy.connect(self)
    for l in segment.dovetails + segment.containments:
      lc = l.clone()
      if lc.from_segment == segment.name:
        lc.from_segment = clone_name
      if lc.to_segment == segment.name:
        lc.to_segment = clone_name
      lc.connect(self)

  LINKS_DISTRIBUTION_POLICY = ["off", "auto", "equal", "L", "R"]
  '''Allowed values for the links_distribution_policy option'''

  def _select_distribute_end(self, links_distribution_policy,
                             segment_name, factor):
    if links_distribution_policy not in self.LINKS_DISTRIBUTION_POLICY:
      raise gfa.ArgumentError("Unknown links distribution policy {}\n".format(\
          links_distribution_policy)+"accepted values are: {}".format(\
          ", ".join(self.LINKS_DISTRIBUTION_POLICY)))
    if links_distribution_policy == "off":
      return None
    if links_distribution_policy in ["L", "R"]:
      return links_distribution_policy
    else:
      s = self.segment(segment_name)
      esize = len(s.dovetails_of_end("R"))
      bsize = len(s.dovetails_of_end("L"))
      return self._auto_select_distribute_end(factor, bsize, esize,
                                       links_distribution_policy == "equal")

  # (keep separate for testing)
  # @tested_in unit_multiplication
  @staticmethod
  def _auto_select_distribute_end(factor, bsize, esize, equal_only):
    if esize == factor:
      return "R"
    elif bsize == factor:
      return "L"
    elif equal_only:
      return None
    elif esize < 2:
      if bsize < 2:
        return None
      else:
        return "L"
    elif bsize < 2:
      return "R"
    elif esize < factor:
      if bsize <= esize:
        return "R"
      elif bsize < factor:
        return "L"
      else:
        return "R"
    elif bsize < factor:
      return "L"
    elif bsize <= esize:
      return "L"
    else:
      return "R"

  def _distribute_links(self, links_distribution_policy, segment_name,
                        copy_names, factor):
    if factor < 2:
      return
    end_type = self._select_distribute_end(links_distribution_policy,
                                           segment_name, factor)
    if end_type is None:
      return
    et_links = self.segment(segment_name).dovetails_of_end(end_type)
    diff = max([len(et_links)-factor, 0])
    links_signatures = list([repr(l.other_end(gfapy.SegmentEnd(segment_name, \
                          end_type))) for l in et_links])
    for i, sn in enumerate([segment_name]+copy_names):
      to_keep = links_signatures[i:i+diff+1]
      links = self.segment(sn).dovetails_of_end(end_type).copy()
      for l in links:
        l_sig = repr(l.other_end(gfapy.SegmentEnd(sn, end_type)))
        if l_sig not in to_keep:
          l.disconnect()

  def _segment_and_segment_name(self, segment_or_segment_name):
    if isinstance(segment_or_segment_name, gfapy.Line):
      s = segment_or_segment_name
      sn = segment_or_segment_name.name
    else:
      s = self.segment(segment_or_segment_name)
      sn = segment_or_segment_name
    return s, sn



class Error(Exception):
  """Parent class for library-specific errors"""
  pass

class VersionError(Error):
  """Unknown/wrong version of the specification"""
  pass

class RuntimeError(Error):
  """The user tried to do something not allowed"""
  pass

class ValueError(Error):
  """An object has the right type/form, but an invalid content

  e.g. number out-of-range; string/array too big/small
  """
  pass

class FormatError(Error):
  """The format of an object is invalid

  e.g. a line contains too many/few fields;
       a tagname has the wrong format
  """
  pass

class TypeError(Error):
  """A wrong type has been used or specified

  e.g. a field contains an array instead of an integer;
       an invalid record type or datatype is found by parsing
  """
  pass

class ArgumentError(Error):
  """The argument of a method has the wrong type"""
  pass

class NotUniqueError(Error):
  """An element which should have been unique is not unique

  e.g. a tag name is duplicated in a line; a duplicated record ID is found
  """
  pass

class InconsistencyError(Error):
  """Contradictory information has been provided

  e.g. GFA1 segment LN and sequence length differ;
       a GFA2-only record is added to a GFA1 file
  """
  pass

class NotFoundError(Error):
  """An element which has been required is not found

  e.g. a tag or record which is required is not found
  """
  pass

class AssertionError(Error):
  """An assertion has failed

  An error of this kind indicates a probable bug.
  """
  pass



import gfapy
import unittest

class TestAPILinesFinders(unittest.TestCase):

  l_gfa1_a = ["S\t1\t*",
              "S\t2\t*",
              "S\t3\t*",
              "S\t4\tCGAT",
              "L\t1\t+\t2\t+\t*",
              "L\t1\t-\t3\t+\t*",
              "C\t1\t-\t4\t-\t1\t*",
              "P\tp1\t1+,2+\t*"]
  l_gfa1 = [gfapy.Line(x) for x in l_gfa1_a]
  l_gfa2_a = ["S\t1\t100\t*",
              "S\t2\t110\t*",
              "E\te1\t1+\t2-\t0\t100$\t10\t110$\t*",
              "G\tg1\t1-\t2-\t1000\t*",
              "O\to1\t1+ 2-",
              "U\tu1\t1 e1",
              "F\t1\tread1-\t0\t10\t102\t122\t*",
              "F\t1\tread1-\t30\t100$\t180\t255\t*",
              "F\t2\tread1-\t40\t50\t52\t64\t*",
              "X\tx1\txx:Z:A",
              "X\tx2",
              "G\t*\t1+\t2+\t2000\t*"]
  l_gfa2 = [gfapy.Line(x) for x in l_gfa2_a]
  gfa1 = gfapy.Gfa(l_gfa1)
  gfa2 = gfapy.Gfa(l_gfa2)

  def test_segment_gfa1(self):
    # existing name as argument
    self.assertEqual(TestAPILinesFinders.l_gfa1[0],TestAPILinesFinders.gfa1.segment("1"))
    self.assertEqual(TestAPILinesFinders.l_gfa1[0],TestAPILinesFinders.gfa1.try_get_segment("1"))
    # not existing name as argument
    self.assertEqual(None,TestAPILinesFinders.gfa1.segment("0"))
    with self.assertRaises(gfapy.NotFoundError): TestAPILinesFinders.gfa1.try_get_segment("0")
    # line as argument
    self.assertEqual(TestAPILinesFinders.l_gfa1[0],TestAPILinesFinders.gfa1.segment(TestAPILinesFinders.l_gfa1[0]))
    self.assertEqual(TestAPILinesFinders.l_gfa1[0],TestAPILinesFinders.gfa1.try_get_segment(TestAPILinesFinders.l_gfa1[0]))
    # connection to rgfa is not checked if argument is line
    self.assertEqual(TestAPILinesFinders.l_gfa2[0],TestAPILinesFinders.gfa1.segment(TestAPILinesFinders.l_gfa2[0]))
    self.assertEqual(TestAPILinesFinders.l_gfa2[0],TestAPILinesFinders.gfa1.try_get_segment(TestAPILinesFinders.l_gfa2[0]))

  def test_segment_gfa2(self):
    # existing name as argument
    self.assertEqual(TestAPILinesFinders.l_gfa2[0],TestAPILinesFinders.gfa2.segment("1"))
    self.assertEqual(TestAPILinesFinders.l_gfa2[0],TestAPILinesFinders.gfa2.try_get_segment("1"))
    # not existing name as argument
    self.assertEqual(None,TestAPILinesFinders.gfa2.segment("0"))
    with self.assertRaises(gfapy.NotFoundError): TestAPILinesFinders.gfa2.try_get_segment("0")
    # line as argument
    self.assertEqual(TestAPILinesFinders.l_gfa2[0],TestAPILinesFinders.gfa2.segment(TestAPILinesFinders.l_gfa2[0]))
    self.assertEqual(TestAPILinesFinders.l_gfa2[0],TestAPILinesFinders.gfa2.try_get_segment(TestAPILinesFinders.l_gfa2[0]))
    # connection to rgfa is not checked if argument is line
    self.assertEqual(TestAPILinesFinders.l_gfa1[0],TestAPILinesFinders.gfa2.segment(TestAPILinesFinders.l_gfa1[0]))
    self.assertEqual(TestAPILinesFinders.l_gfa1[0],TestAPILinesFinders.gfa2.try_get_segment(TestAPILinesFinders.l_gfa1[0]))

  def test_line_gfa1(self):
    # segment name as argument
    self.assertEqual(TestAPILinesFinders.l_gfa1[0],TestAPILinesFinders.gfa1.line("1"))
    self.assertEqual(TestAPILinesFinders.l_gfa1[0],TestAPILinesFinders.gfa1.try_get_line("1"))
    # path name as argument
    self.assertEqual(TestAPILinesFinders.l_gfa1[7],TestAPILinesFinders.gfa1.line("p1"))
    self.assertEqual(TestAPILinesFinders.l_gfa1[7],TestAPILinesFinders.gfa1.try_get_line("p1"))
    # not existing name as argument
    self.assertEqual(None,TestAPILinesFinders.gfa1.line("0"))
    with self.assertRaises(gfapy.NotFoundError): TestAPILinesFinders.gfa1.try_get_line("0")
    # line as argument
    self.assertEqual(TestAPILinesFinders.l_gfa1[0],TestAPILinesFinders.gfa1.line(TestAPILinesFinders.l_gfa1[0]))
    self.assertEqual(TestAPILinesFinders.l_gfa1[0],TestAPILinesFinders.gfa1.try_get_line(TestAPILinesFinders.l_gfa1[0]))
    # connection to rgfa is not checked if argument is line
    self.assertEqual(TestAPILinesFinders.l_gfa2[0],TestAPILinesFinders.gfa1.line(TestAPILinesFinders.l_gfa2[0]))
    self.assertEqual(TestAPILinesFinders.l_gfa2[0],TestAPILinesFinders.gfa1.try_get_line(TestAPILinesFinders.l_gfa2[0]))

  def test_line_gfa2(self):
    # segment name as argument
    self.assertEqual(TestAPILinesFinders.l_gfa2[0],TestAPILinesFinders.gfa2.line("1"))
    self.assertEqual(TestAPILinesFinders.l_gfa2[0],TestAPILinesFinders.gfa2.try_get_line("1"))
    # edge name as argument
    self.assertEqual(TestAPILinesFinders.l_gfa2[2],TestAPILinesFinders.gfa2.line("e1"))
    self.assertEqual(TestAPILinesFinders.l_gfa2[2],TestAPILinesFinders.gfa2.try_get_line("e1"))
    # gap name as argument
    self.assertEqual(TestAPILinesFinders.l_gfa2[3],TestAPILinesFinders.gfa2.line("g1"))
    self.assertEqual(TestAPILinesFinders.l_gfa2[3],TestAPILinesFinders.gfa2.try_get_line("g1"))
    # path name as argument
    self.assertEqual(TestAPILinesFinders.l_gfa2[4],TestAPILinesFinders.gfa2.line("o1"))
    self.assertEqual(TestAPILinesFinders.l_gfa2[4],TestAPILinesFinders.gfa2.try_get_line("o1"))
    # set name as argument
    self.assertEqual(TestAPILinesFinders.l_gfa2[5],TestAPILinesFinders.gfa2.line("u1"))
    self.assertEqual(TestAPILinesFinders.l_gfa2[5],TestAPILinesFinders.gfa2.try_get_line("u1"))
    # not existing name as argument
    self.assertIsNone(TestAPILinesFinders.gfa2.line("0"))
    with self.assertRaises(gfapy.NotFoundError): TestAPILinesFinders.gfa2.try_get_line("0")
    # line as argument
    self.assertEqual(TestAPILinesFinders.l_gfa2[0],TestAPILinesFinders.gfa2.line(TestAPILinesFinders.l_gfa2[0]))
    self.assertEqual(TestAPILinesFinders.l_gfa2[0],TestAPILinesFinders.gfa2.try_get_line(TestAPILinesFinders.l_gfa2[0]))
    # connection to rgfa is not checked if argument is line
    self.assertEqual(TestAPILinesFinders.l_gfa1[0],TestAPILinesFinders.gfa2.line(TestAPILinesFinders.l_gfa1[0]))
    self.assertEqual(TestAPILinesFinders.l_gfa1[0],TestAPILinesFinders.gfa2.try_get_line(TestAPILinesFinders.l_gfa1[0]))

  def test_fragments_for_external(self):
    self.assertEqual(TestAPILinesFinders.l_gfa2[6:9], TestAPILinesFinders.gfa2.fragments_for_external("read1"))
    self.assertEqual([], TestAPILinesFinders.gfa2.fragments_for_external("read2"))

  def test_select_by_hash_gfa1(self):
    # search segments
    self.assertEqual(set(TestAPILinesFinders.l_gfa1_a[0:4]),
        set([str(x) for x in TestAPILinesFinders.gfa1.select({"record_type":"S",
                                                "sequence":"CGAT"})]))
    self.assertEqual(TestAPILinesFinders.l_gfa1[0:1], TestAPILinesFinders.gfa1.select({"record_type":"S",
                                                "name":"1"}))
    # search links
    self.assertEqual(TestAPILinesFinders.l_gfa1[4:5], TestAPILinesFinders.gfa1.select({"record_type":"L",
                                                "from":"1",
                                                "from_orient":"+"}))
    # search containments
    self.assertEqual(TestAPILinesFinders.l_gfa1[6:7], TestAPILinesFinders.gfa1.select({"record_type":"C",
                                                "from":"1",
                                                "pos":1}))
    # search paths
    self.assertEqual(TestAPILinesFinders.l_gfa1[7:8], TestAPILinesFinders.gfa1.select({"record_type":"P",
                                                "segment_names":"1+,2+"}))
    # no record type specified
    self.assertEqual(TestAPILinesFinders.l_gfa1[0:1], TestAPILinesFinders.gfa1.select({"name":"1"}))
    self.assertEqual(TestAPILinesFinders.l_gfa1[4:7], TestAPILinesFinders.gfa1.select({"from":"1"}))
    # reference as value
    self.assertEqual(TestAPILinesFinders.l_gfa1[4:7], TestAPILinesFinders.gfa1.select({"from":TestAPILinesFinders.l_gfa1[0]}))
    # placeholder is equal to any value
    self.assertEqual(set(TestAPILinesFinders.l_gfa1_a[0:3]),
        set([str(x) for x in TestAPILinesFinders.gfa1.select({"sequence":"ACC"})]))

  def test_select_by_line_gfa1(self):
    for i in range(len(TestAPILinesFinders.l_gfa1)):
      self.assertEqual(TestAPILinesFinders.l_gfa1[i:i+1],
          TestAPILinesFinders.gfa1.select(TestAPILinesFinders.l_gfa1[i]))

  def test_select_by_hash_gfa2(self):
    # search segments
    self.assertEqual(set(TestAPILinesFinders.l_gfa2_a[0:2]),
        set([str(x) for x in TestAPILinesFinders.gfa2.select({"record_type":"S",
                                                "sequence":"CGAT"})]))
    self.assertEqual(TestAPILinesFinders.l_gfa2[1:2], TestAPILinesFinders.gfa2.select({"record_type":"S",
                                                "slen":110}))
    # search edges
    self.assertEqual(TestAPILinesFinders.l_gfa2[2:3], TestAPILinesFinders.gfa2.select({"record_type":"E",
                                                "sid1":gfapy.OrientedLine("1","+")}))
    # search gaps
    self.assertEqual(TestAPILinesFinders.l_gfa2[3:4], TestAPILinesFinders.gfa2.select({"record_type":"G",
                                                "sid1":gfapy.OrientedLine("1","-")}))
    self.assertEqual(TestAPILinesFinders.l_gfa2[11:12], TestAPILinesFinders.gfa2.select({"record_type":"G",
                                                "disp":2000}))
    # search paths
    self.assertEqual(TestAPILinesFinders.l_gfa2[4:5], TestAPILinesFinders.gfa2.select({"record_type":"O",
                                                "items":"1+ 2-"}))
    # search sets
    self.assertEqual(TestAPILinesFinders.l_gfa2[5:6], TestAPILinesFinders.gfa2.select({"record_type":"U",
                                                "name":"u1"}))
    # search fragments
    self.assertEqual(TestAPILinesFinders.l_gfa2[6:9], TestAPILinesFinders.gfa2.select({"record_type":"F",
                                                "external":"read1-"}))
    # search custom records
    self.assertEqual(TestAPILinesFinders.l_gfa2[9:10], TestAPILinesFinders.gfa2.select({"record_type":"X",
                                                "xx":"A"}))

  def test_select_by_line_gfa2(self):
    for i in range(len(TestAPILinesFinders.l_gfa2)):
      self.assertEqual(TestAPILinesFinders.l_gfa2[i:i+1],
          TestAPILinesFinders.gfa2.select(TestAPILinesFinders.l_gfa2[i]))




import gfapy
import re

def unsafe_decode(string):
  return gfapy.ByteArray(string)

def decode(string):
  return gfapy.ByteArray(string)

def validate_encoded(string):
  if not re.match(r"^[0-9A-F]+$", string):
    raise gfapy.FormatError(
      "{} is not a valid hex string\n".format(repr(string))+
      "(it does not match the regular expression [0-9A-F]+)")

def validate_decoded(byte_array):
  return byte_array.validate()

def unsafe_encode(obj):
  if isinstance(obj, gfapy.ByteArray):
    return str(obj)
  if isinstance(obj, list):
    return str(ByteArray(obj))
  elif isinstance(obj, str):
    return obj
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: str, list, gfapy.ByteArray)")

def encode(obj):
  if isinstance(obj, gfapy.ByteArray):
    return str(obj)
  elif isinstance(obj, list):
    return str(ByteArray(obj))
  elif isinstance(obj, str):
    validate_encoded(obj)
    return obj
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: str, list, gfapy.ByteArray)")



class SameID:

  def _process_not_unique(self, previous):
    self._gfa = previous.gfa
    self._initialize_references()
    cur_items = self.get("items")
    self._substitute_virtual_line(previous)
    self._set_existing_field("items", self.get("items") + cur_items, 
                            set_reference = True)
    self._import_tags_of_previous_group_definition(previous)
    return None

  def _import_tags_of_previous_group_definition(self, previous):
    for tag in previous.tagnames:
      prv = previous.get(tag)
      cur = self.get(tag)
      if cur:
        if cur != prv:
          raise gfapy.NotUniqueError(
            "Same tag defined differently in "+
            "multiple group lines with same ID\n"+
            "Previous tag definition: {}\n".format(prv)+
            "New tag definition: {}\n".format(cur)+
            "Group ID: {}".format(self.name))
      else:
        self.set(tag, prv)



import gfapy

class Other:

  def other_oriented_segment(self, oriented_segment, tolerant = False):
    """
    Parameters
    ----------
    oriented_segment : gfapy.OrientedLine
      One of the two oriented segments of the line.

    Returns
    -------
    gfapy.OrientedLine
      The other oriented segment.

    Raises
    ------
    gfapy.NotFoundError
      If segment_end is not a segment end of the line.
    """
    if (self.oriented_from == oriented_segment):
      return self.oriented_to
    elif (self.oriented_to == oriented_segment):
      return self.oriented_from
    elif tolerant:
      return None
    else:
      raise gfapy.NotFoundError(
        "Oriented segment '{}' not found\n".format(repr(oriented_segment))+
        "Line: {}".format(self))

  def other(self, segment, tolerant = False):
    """
    The other segment of a connection line.

    Parameters
    ----------
    segment : gfapy.line.segment.GFA1 or str
      Segment name or instance.

    Raises
    ------
    gfapy.NotFoundError
      If segment is not involved in the connection.

    Returns
    -------
    str
      The name or instance of the other segment of the connection.
      If circular, then **segment**.
    """
    segment_name = str(segment)
    if segment_name == str(self.from_segment):
      return self.to
    elif segment_name == str(self.to_segment):
      return self.from_segment
    elif tolerant:
      return None
    else:
      raise gfapy.NotFoundError(
        "Line {} does not involve segment {}".format(self, segment_name))



import unittest
import gfapy

class TestApiTags(unittest.TestCase):

  def test_predefined_tags(self):
    # correct type:
    gfapy.line.Header(["H", "VN:Z:1"], vlevel=3) # nothing raised
    # custom tags with the same letters as predefined tags but lower case
    gfapy.line.Header(["H", "vn:i:1"], vlevel=3) # nothing raised
    # wrong type
    gfapy.line.Header(["H", "VN:i:1"], vlevel=0) # nothing raised
    for level in [1,2,3]:
      self.assertRaises(gfapy.TypeError,
        gfapy.line.Header, ["H", "VN:i:1"], vlevel=level)

  def test_custom_tags(self):
    for version in ["gfa1","gfa2"]:
      for tagname in ["ZZ","Z1","Zz","zz"]:
        for level in [0,1,2,3]:
          tag = "{}:Z:1".format(tagname)
          gfapy.line.Header(["H", tag], version=version, vlevel=level) # nothing raised
          gfapy.Line("H\t"+tag, version=version, vlevel=level) # nothing raised
          gfapy.Gfa("H\t"+tag, version=version, vlevel=level) # nothing raised

  def test_wrong_tag_format(self):
    self.assertRaises(gfapy.FormatError, gfapy.line.Header, ["H", "VN i:1"])
    self.assertRaises(gfapy.FormatError, gfapy.line.Header, ["H", "vna:i:1"])
    self.assertRaises(gfapy.FormatError, gfapy.line.Header, ["H", "VN:ZZ:1"])
    # the content can include :, so four : are e.g. not an error
    self.assertEqual("1:1:1", gfapy.line.Header(["H", "VN:Z:1:1:1"]).VN)

  def test_wrong_tag_data(self):
    # validation level 0
    # - some wrong data passes through
    gfapy.line.Header(["H", "zz:B:i,1,1,A"], vlevel=0) # nothing raised
    gfapy.line.Header(["H", "zz:Z:i,\t1,1,A"], vlevel=0) # nothing raised
    # - some errors are catched
    self.assertRaises(gfapy.FormatError, gfapy.line.Header, ["H", "zz:i:1A"], vlevel=0)
    # level > 0, wrong data is catched
    for level in [1,2,3]:
      self.assertRaises(gfapy.ValueError,
        gfapy.line.Header,["H", "zz:B:i,1,1,A"],vlevel=level)
      self.assertRaises(gfapy.FormatError,
        gfapy.line.Header,["H", "zz:i:1A"],vlevel=level)

  def test_duplicate_tag(self):
    for version in ["gfa1","gfa2"]:
      gfapy.line.Header(["H", "zz:i:1", "VN:Z:1", "zz:i:2"],
                              version=version, vlevel=0) # nothing raised
      gfapy.Line("H\tzz:i:1\tVN:Z:0\tzz:i:2",version=version,
                                               vlevel=0) # nothing raised
      gfapy.Line("H\tzz:i:1\tVN:Z:0\tzz:i:2",version=version,
                                          vlevel=0) # nothing raised
      for level in [1,2,3]:
        self.assertRaises(gfapy.NotUniqueError, gfapy.line.Header,
          ["H", "zz:i:1", "VN:Z:0", "zz:i:2"], version=version, vlevel=level)
        self.assertRaises(gfapy.NotUniqueError, gfapy.Line,
            "H\tzz:i:1\tVN:Z:0\tzz:i:2",version=version, vlevel=level)
        self.assertRaises(gfapy.NotUniqueError, gfapy.Gfa,
            "H\tzz:i:1\tVN:Z:#{version}\tzz:i:2", version=version, vlevel=level)

  def test_validate_field(self):
    l = gfapy.line.Header(["H", "zz:i:1", "VN:Z:1.0"], version="gfa1", vlevel=0)
    l.zz = "x"
    self.assertRaises(gfapy.FormatError, l.validate_field, "zz")
    l.set_datatype("zz", "Z")
    l.validate_field("zz") # nothing raised

  def test_validate(self):
    # wrong tag value
    l = gfapy.line.Header(["H", "zz:i:1", "VN:Z:1.0"], version="gfa1", vlevel=0)
    l.zz = "x"
    self.assertRaises(gfapy.FormatError, l.validate)
    # wrong predefined tag datatype
    l = gfapy.line.Header(["H", "zz:i:1", "VN:i:1"], version="gfa1", vlevel=0)
    self.assertRaises(gfapy.TypeError, l.validate)

  # test tags for get/set tests:
  # - KC -> predefined, set
  # - RC -> predefined, not set;
  # - xx -> custom set
  # - zz -> custom not set
  # - XX -> custom, not set, upper case

  def test_get_tag_content(self):
    for version in ["gfa1","gfa2"]:
      for level in [0,1,2,3]:
        l = gfapy.Line(["S", "12","*","xx:f:1.3","KC:i:10"], vlevel=level)
        # tagnames
        self.assertEqual(sorted(["xx", "KC"]), sorted(l.tagnames))
        # test presence of tag
        assert(l.KC)
        assert(not l.RC)
        assert(not l.XX)
        assert(l.xx)
        assert(not l.zz)
        # tagname as attribute
        self.assertEqual(10, l.KC)
        self.assertEqual(None, l.RC)
        self.assertEqual(None, l.XX)
        self.assertEqual(1.3, l.xx)
        self.assertEqual(None, l.zz)
        # get(tagname)
        self.assertEqual(10, l.get("KC"))
        self.assertEqual(None, l.get("RC"))
        self.assertEqual(None, l.get("XX"))
        self.assertEqual(1.3, l.get("xx"))
        self.assertEqual(None, l.get("zz"))
        # try_get_<tagname>()
        self.assertEqual(10, l.try_get_KC())
        self.assertRaises(gfapy.NotFoundError, l.try_get_RC)
        with self.assertRaises(gfapy.NotFoundError):
          l.try_get_XX()
        self.assertEqual(1.3, l.try_get_xx())
        with self.assertRaises(gfapy.NotFoundError):
          l.try_get_zz()
        # try_get(tagname)
        self.assertEqual(10, l.try_get("KC"))
        self.assertRaises(gfapy.NotFoundError, l.try_get, "RC")
        self.assertRaises(gfapy.NotFoundError, l.try_get, "XX")
        self.assertEqual(1.3, l.try_get("xx"))
        self.assertRaises(gfapy.NotFoundError, l.try_get, "zz")
        # get_datatype(tagname)
        self.assertEqual("i", l.get_datatype("KC"))
        self.assertEqual("i", l.get_datatype("RC"))
        self.assertEqual(None, l.get_datatype("XX"))
        self.assertEqual("f", l.get_datatype("xx"))
        self.assertEqual(None, l.get_datatype("zz"))
        # field_to_s(tagname, tag=False)
        self.assertEqual("10", l.field_to_s("KC"))
        self.assertRaises(gfapy.NotFoundError, l.field_to_s, "RC")
        self.assertRaises(gfapy.NotFoundError, l.field_to_s, "XX")
        self.assertEqual("1.3", l.field_to_s("xx"))
        self.assertRaises(gfapy.NotFoundError, l.field_to_s, "zz")
        # field_to_s(tagname, tag=True)
        self.assertEqual("KC:i:10", l.field_to_s("KC", tag=True))
        self.assertEqual("xx:f:1.3", l.field_to_s("xx", tag=True))

  def test_set_tag_content(self):
    for version in ["gfa1","gfa2"]:
      for level in [0,1,2,3]:
        l = gfapy.Line(["S", "12","*","xx:f:13","KC:i:10"], vlevel=level)
        # set tag content, fieldname methods
        l.KC = 12 # nothing raised; self.assertEqual(12, l.KC)
        l.RC = 12 # nothing raised; self.assertEqual(12, l.RC)
        l.xx = 1.2 # nothing raised; self.assertEqual(1.2, l.xx)
        l.zz = 1.2 # nothing raised; self.assertEqual(1.2, l.zz)
        # set tag content, set()
        l.set("KC", 14) # nothing raised; self.assertEqual(14, l.KC)
        l.set("RC", 14) # nothing raised; self.assertEqual(14, l.RC)
        l.set("xx", 1.4) # nothing raised; self.assertEqual(1.4, l.xx)
        l.set("zz", 1.4) # nothing raised; self.assertEqual(1.4, l.zz)
        # set datatype for predefined field
        self.assertRaises(gfapy.RuntimeError, l.set_datatype, "KC","Z")
        self.assertRaises(gfapy.RuntimeError, l.set_datatype, "RC","Z")
        # set datatype for non-existing custom tag
        l.set_datatype("zz", "i") # nothing raised
        l.set_datatype("XX", "Z") # nothing raised
        # change datatype for existing custom tag
        l.xx = 1.1 # nothing raised
        l.xx = "1.1" # nothing raised
        if level == 2:
          l.xx = "1A" # nothing raised
          with self.assertRaises(gfapy.FormatError):
            str(l)
        elif level == 3:
          with self.assertRaises(gfapy.FormatError):
            l.xx = "1A"
        l.set_datatype("xx", "Z"); l.xx = "1A"  # nothing raised
        # unknown datatype
        self.assertRaises(gfapy.ArgumentError, l.set_datatype, "xx", "P")

  def test_delete_tag(self):
    for version in ["gfa1","gfa2"]:
      for level in [0,1,2,3]:
        l = gfapy.Line(["S", "12","*","xx:f:13","KC:i:10"], vlevel=level)
        # delete method
        l.delete("KC") # nothing raised
        self.assertEqual(None, l.KC)
        self.assertEqual(["xx"], l.tagnames)
        l.delete("RC") # nothing raised
        l.delete("XX") # nothing raised
        l.delete("xx") # nothing raised
        self.assertEqual([], l.tagnames)
        l.delete("zz") # nothing raised
        l = gfapy.Line(["S", "12","*","xx:f:13","KC:i:10"], vlevel=level)
        # set to None
        l.set("KC",None)  # nothing raised
        self.assertEqual(None, l.KC)
        self.assertEqual(["xx"], l.tagnames)
        l.set("RC",None)  # nothing raised
        l.set("XX",None)  # nothing raised
        l.set("xx",None)  # nothing raised
        self.assertEqual([], l.tagnames)
        l.set("zz",None)  # nothing raised

  def test_datatype_to_python_objects(self):
    l = gfapy.line.Header(["H", "a1:A:1", "z1:Z:hallo",
                                "b1:B:c,12,12", "b2:B:f,1E-2,3.0,3",
                                "h1:H:00A1",
                                "j1:J:[12,\"a\"]", "j2:J:{\"a\":1,\"b\":[2,3]}",
                                "f1:f:-1.23E-04", "i1:i:-123"])
    self.assertEqual(str, l.a1.__class__)
    self.assertEqual(str, l.z1.__class__)
    self.assertEqual(gfapy.NumericArray, l.b1.__class__)
    self.assertEqual(gfapy.NumericArray, l.b2.__class__)
    self.assertEqual(gfapy.ByteArray, l.h1.__class__)
    self.assertEqual(list, l.j1.__class__)
    self.assertEqual(dict, l.j2.__class__)
    self.assertEqual(int, l.i1.__class__)
    self.assertEqual(float, l.f1.__class__)


  def test_python_object_to_datatype(self):
    l = gfapy.line.Header(["H"])
    # String
    l.zz="1"  # nothing raised
    self.assertEqual("1", l.zz)
    self.assertEqual("Z", l.get_datatype("zz"))
    self.assertEqual("1", l.field_to_s("zz"))
    self.assertEqual("1", gfapy.Line(str(l)).zz)
    # Integer
    l.ii=1  # nothing raised
    self.assertEqual(1, l.ii)
    self.assertEqual("i", l.get_datatype("ii"))
    self.assertEqual("1", l.field_to_s("ii"))
    self.assertEqual(1, gfapy.Line(str(l)).ii)
    # Float
    l.ff=1.0  # nothing raised
    self.assertEqual(1.0, l.ff)
    self.assertEqual("f", l.get_datatype("ff"))
    self.assertEqual("1.0", l.field_to_s("ff"))
    self.assertEqual(1.0, gfapy.Line(str(l)).ff)
    # Array: all floats
    l.af=[1.0,1.0]  # nothing raised
    self.assertEqual([1.0,1.0], l.af)
    self.assertEqual("B", l.get_datatype("af"))
    self.assertEqual("f,1.0,1.0", l.field_to_s("af"))
    self.assertEqual([1.0,1.0], gfapy.Line(str(l)).af)
    # Array: all integers
    l.ai=[1,1]  # nothing raised
    self.assertEqual([1,1], l.ai)
    self.assertEqual("B", l.get_datatype("ai"))
    self.assertEqual("C,1,1", l.field_to_s("ai"))
    self.assertEqual([1,1], gfapy.Line(str(l)).ai)
    # Array: anything else
    l.aa=[1,1.0,"X"]  # nothing raised
    self.assertEqual([1,1.0,"X"], l.aa)
    self.assertEqual("J", l.get_datatype("aa"))
    self.assertEqual('[1, 1.0, "X"]', l.field_to_s("aa"))
    self.assertEqual([1,1.0,"X"], gfapy.Line(str(l)).aa)
    # Hash
    l.hh={"a":1.0, "b":1}  # nothing raised
    self.assertEqual({"a":1.0,"b":1}, l.hh)
    self.assertEqual("J", l.get_datatype("hh"))
    try:
      self.assertEqual('{"a": 1.0, "b": 1}', l.field_to_s("hh"))
    except:
      self.assertEqual('{"b": 1, "a": 1.0}', l.field_to_s("hh"))
    self.assertEqual({"a":1.0,"b":1}, gfapy.Line(str(l)).hh)
    # gfapy.ByteArray
    l.ba=gfapy.ByteArray([0,255])  # nothing raised
    self.assertEqual(gfapy.ByteArray([0,255]), l.ba)
    self.assertEqual("H", l.get_datatype("ba"))
    self.assertEqual('00FF', l.field_to_s("ba"))
    self.assertEqual(gfapy.ByteArray([0,255]), gfapy.Line(str(l)).ba)

  def test_byte_arrays(self):
    # creation:, from array, from string
    a = gfapy.ByteArray([1,2,3,4,5]) # nothing raised
    b = gfapy.ByteArray([1,2,3,4,5]) # nothing raised
    self.assertEqual(a, b)
    c = gfapy.ByteArray("12ACF4AA601C1F") # nothing raised
    self.assertEqual(gfapy.ByteArray([18, 172, 244, 170, 96, 28, 31]), c)
    # validation
    a.validate()  # nothing raised
    with self.assertRaises(gfapy.ValueError):
      gfapy.ByteArray([1,2,3,4,356])
    with self.assertRaises(gfapy.FormatError):
      gfapy.ByteArray("12ACF4AA601C1")
    with self.assertRaises(gfapy.FormatError):
      gfapy.ByteArray("")
    with self.assertRaises(gfapy.FormatError):
      gfapy.ByteArray("12ACG4AA601C1F")
    # to string
    a = gfapy.ByteArray([18, 172, 244, 170, 96, 28, 31])
    self.assertEqual("12ACF4AA601C1F", str(a))
    a = list(a)
    a[2] = 280
    with self.assertRaises(gfapy.ValueError):
      a = gfapy.ByteArray(a)

  def test_numeric_arrays(self):
    # creation:, from array, from string
    a = gfapy.NumericArray([1,2,3,4,5]) # nothing raised
    b = gfapy.NumericArray([1,2,3,4,5]) # nothing raised
    self.assertEqual(a, b)
    c = gfapy.NumericArray.from_string("i,1,2,3,4,5") # nothing raised
    self.assertEqual(gfapy.NumericArray([1, 2, 3, 4, 5]), c)
    # validation
    a.validate()  # nothing raised
    gfapy.NumericArray([1,2,3,4,356]).validate()  # nothing raised
    self.assertRaises(gfapy.ValueError,
      gfapy.NumericArray([1,2.0,3,4,356]).validate)
    self.assertRaises(gfapy.ValueError,
      gfapy.NumericArray([1.0,2.0,3,4,356]).validate)
    self.assertRaises(gfapy.ValueError,
      gfapy.NumericArray([1,"x",3,4,356]).validate)
    with self.assertRaises(gfapy.ValueError):
        a = gfapy.NumericArray.from_string("i,1,X,2")
    with self.assertRaises(gfapy.FormatError):
        a = gfapy.NumericArray.from_string("")
    with self.assertRaises(gfapy.FormatError):
        a = gfapy.NumericArray.from_string("i,1,2,")
    with self.assertRaises(gfapy.TypeError):
        a = gfapy.NumericArray.from_string("x,1,2")
    # to string
    a = gfapy.NumericArray([18, 72, 244, 70, 96, 38, 31])
    self.assertEqual("C", a.compute_subtype())
    self.assertEqual("C,18,72,244,70,96,38,31", str(a))
    a[2] = -2
    self.assertEqual("c", a.compute_subtype())
    self.assertEqual("c,18,72,-2,70,96,38,31", str(a))
    a[2] = 280
    self.assertEqual("S", a.compute_subtype())
    self.assertEqual("S,18,72,280,70,96,38,31", str(a))
    a[2] = -280
    self.assertEqual("s", a.compute_subtype())
    self.assertEqual("s,18,72,-280,70,96,38,31", str(a))
    a[2] = 280000
    self.assertEqual("I", a.compute_subtype())
    self.assertEqual("I,18,72,280000,70,96,38,31", str(a))
    a[2] = -280000
    self.assertEqual("i", a.compute_subtype())
    self.assertEqual("i,18,72,-280000,70,96,38,31", str(a))
    a = gfapy.NumericArray([float(x) for x in a])
    self.assertEqual("f", a.compute_subtype())
    self.assertEqual("f,18.0,72.0,-280000.0,70.0,96.0,38.0,31.0", str(a))



VERSIONS = ["gfa1", "gfa2"]
DIALECTS = ["rgfa", "standard"]
from gfapy.error import *
from gfapy.placeholder import Placeholder
from gfapy.placeholder import is_placeholder
from gfapy.byte_array import ByteArray
from gfapy.field_array import FieldArray
from gfapy.alignment import Alignment
from gfapy.alignment.cigar import CIGAR
from gfapy.alignment.placeholder import AlignmentPlaceholder
from gfapy.alignment.trace import Trace
from gfapy.numeric_array import NumericArray
from gfapy.lastpos import LastPos
from gfapy.lastpos import isfirstpos, islastpos, posvalue
from gfapy.symbol_invert import invert
from gfapy.field import Field
from gfapy.line import Line
from gfapy.logger import Logger
from gfapy.segment_end_path import SegmentEndsPath
from gfapy.segment_end import *
from gfapy.oriented_line import OrientedLine
from gfapy.lines import Lines
from gfapy.graph_operations import GraphOperations
from gfapy.gfa import Gfa
import gfapy.sequence
import gfapy.field



import gfapy

class Alignment:
  """Factory for instances of classes which represent alignments in GFA fields.

  Args:
    initializer (string, list): the alignment field content
    version (str): GFA version, either ``'gfa1'`` or ``'gfa2'`` (default: ``'gfa2'``)
    valid (bool): if ``True``, validation is skipped, when possible (default: ``False``)

  Returns:
    :class:`~gfapy.alignment.cigar.CIGAR`,
    :class:`~gfapy.alignment.trace.Trace`,
    :class:`~gfapy.alignment.placeholder.AlignmentPlaceholder`

  Raises:
    ~gfapy.error.ArgumentError: If more the one positional parameter is used.
    ~gfapy.error.FormatError: If the ``initializer`` string/list is invalid.
    ~gfapy.error.VersionError: If ``version`` is invalid, or ``initializer`` is
        incompatible with the ``version``.

  Examples:
    >>> import gfapy
    >>> from gfapy.alignment import Alignment
    >>> Alignment("*")
    gfapy.AlignmentPlaceholder()
    >>> Alignment("12M2I2D")
    gfapy.CIGAR([gfapy.CIGAR.Operation(12,'M'), gfapy.CIGAR.Operation(2,'I'), gfapy.CIGAR.Operation(2,'D')])
    >>> Alignment("12,23,1")
    gfapy.Trace([12,23,1])
    >>> Alignment([])
    gfapy.AlignmentPlaceholder()
    >>> Alignment([gfapy.CIGAR.Operation(12,'M'), gfapy.CIGAR.Operation(2,'I'), gfapy.CIGAR.Operation(2,'D')])
    gfapy.CIGAR([gfapy.CIGAR.Operation(12,'M'), gfapy.CIGAR.Operation(2,'I'), gfapy.CIGAR.Operation(2,'D')])
    >>> Alignment([12,23,1])
    gfapy.Trace([12,23,1])
  """

  def __new__(cls, *args, **kargs):
    """Create an instance of an alignment field class."""
    if args[0] is None or \
        gfapy.is_placeholder(args[0]):
      return gfapy.AlignmentPlaceholder()
    if len(args) > 1:
      raise gfapy.ArgumentError("The Alignment() constructor requires "+
          "a single positional argument, {} found".format(len(args)))
    if isinstance(args[0], gfapy.CIGAR) or \
        isinstance(args[0], gfapy.Trace):
      return args[0]
    if isinstance(args[0], str):
      return Alignment._from_string(*args, **kargs)
    elif isinstance(args[0], list):
      return Alignment._from_list(*args, **kargs)
    else:
      raise gfapy.ArgumentError("Cannot create an alignment "+
          "from an instance of the class {}".format(type(args[0])))

  @classmethod
  def _from_string(cls, string, version = "gfa2", valid = False):
    """
    Parses an alignment field

    Parameters
    ----------
    string : str
      The string to parse.
    version : str
      GFA version (gfa1 or gfa2)
      If *gfa1*, then CIGARs and Placeholders are supported.
      If *gfa2*, also Traces are supported.
      Defaults to *gfa2*.
    valid : bool
      If *True*, the string is guaranteed to be valid, and
      further checks are skipped.
      Defaults to *False*.

    Returns
    -------
    gfapy.CIGAR or gfapy.Trace or gfapy.AlignentPlaceholder

    Raises
    ------
    gfapy.FormatError
      If the content of the field cannot be parsed.
    gfapy.VersionError
      If a wrong value is provided for the version parameter.
    """
    if version != "gfa1" and version != "gfa2":
      raise gfapy.VersionError(
          "Version error: {}".format(repr(version)))
    first = True
    for char in string:
      if first:
        if char.isdigit():
          first = False
          continue
        elif char == "*" and len(string) == 1:
          return gfapy.AlignmentPlaceholder()
      else:
        if char.isdigit():
          continue
        elif char == ",":
          if version == "gfa2":
            t = gfapy.Trace._from_string(string)
            if not valid:
              t.validate()
            return t
          else:
            raise gfapy.FormatError(
                  "Trace alignments are not allowed in GFA1: {}"
                  .format(repr(string)))
        elif char in ["M","I","D","P"] or (char in ["=","X","S","H","N"]
            and version == "gfa1"):
          return gfapy.CIGAR._from_string(string, valid=valid, version=version)
      break
    raise gfapy.FormatError("Alignment field contains invalid data {}"
                            .format(repr(string)))


  @classmethod
  def _from_list(cls, array, version = "gfa2", valid = True):
    """
    Converts an alignment array into a specific list type

    Parameters
    ----------
    array : list
      The alignment array.
    version : str
      GFA version (gfa1 or gfa2)
      If *gfa1*, then CIGARs and Placeholders are supported.
      If *gfa2*, also Traces are supported.
      Defaults to *gfa2*.
    valid : bool
      If *True*, the list is guaranteed to be valid, and
      further checks are skipped.
      Defaults to *False*.

    Returns
    -------
    gfapy.CIGAR or gfapy.Trace
    """
    if version != "gfa1" and version != "gfa2":
      raise gfapy.VersionError(
          "Version error: {}".format(repr(version)))
    if not array:
      return gfapy.AlignmentPlaceholder()
    elif isinstance(array[0], int):
      if version == "gfa2":
        return gfapy.Trace(array)
      else:
        raise gfapy.VersionError(
          "Trace alignments are not allowed in GFA1: {}".format(repr(array)))
    elif isinstance(array[0], gfapy.CIGAR.Operation):
      return gfapy.CIGAR(array)
    else:
      raise gfapy.FormatError(
        "Array does not represent a valid alignment field: {}"
        .format(repr(array)))



import gfapy
import re

def decode(string):
  if string == "*":
    return gfapy.Placeholder()
  else:
    try:
      return int(string)
    except:
      raise gfapy.FormatError("the string does not represent a valid integer")

unsafe_decode = decode

def validate_decoded(obj):
  if isinstance(obj, int) or isinstance(object, gfapy.Placeholder):
    pass
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: int, gfapy.Placeholder)")

def validate_encoded(string):
  if not re.match(r"^(\*|[-+]?[0-9]+)$", string):
    raise gfapy.FormatError(
      "{} does not represent a valid optional integer value\n"
      .format(repr(string))+
      "(it is not * and does not match the regular expression [-+]?[0-9]+)")

def unsafe_encode(obj):
  return str(obj)

def encode(obj):
  if isinstance(obj, str):
    validate_encoded(obj)
    return obj
  elif isinstance(obj, int) or isinstance(obj, gfapy.Placeholder):
    return str(obj)
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: int, str, gfapy.Placeholder)")



class PBubbles:

  def remove_p_bubbles(self):
    '''Removes all p-bubbles in the graph'''
    visited = set()
    for s in self.segments:
      sn = s.name
      if sn in visited:
        continue
      if s.connectivity == (1,1):
        s1 = s.neighbours_of_end("L")[0]
        s2 = s.neighbours_of_end("R")[0]
        n1 = sorted(s1.neighbours, key=lambda s:s.name)
        n2 = sorted(s2.neighbours, key=lambda s:s.name)
        for se in n1:
          visited.add(se[0].name)
        if list(n1) == [os.inverted() for os in n2]:
          self._remove_proven_p_bubble(s1, s2, n1)

  def remove_p_bubble(self, segment_end1, segment_end2,
                      count_tag=None, unit_length=None):
    '''Removes a p-bubble between segment_end1 and segment_end2'''
    if count_tag is None:
      count_tag=self.default["count_tag"]
    if unit_length is None:
      unit_length=self.default["unit_length"]
    s1 = segment(segment_end1.segment)
    s2 = segment(segment_end2.segment)
    et1 = segment(segment_end1.end_type)
    et2 = segment(segment_end2.end_type)
    n1 = sorted(s1.neighbours(et1), key=lambda s:s.name)
    n2 = sorted(s2.neighbours(et2), key=lambda s:s.name)
    assert list(n1) == [os.inverted() for os in n2]
    assert all(se[0].connectivity == (1,1) for se in n1)
    self._remove_proven_p_bubble(segment_end1, segment_end2, n1,
                           count_tag=count_tag, unit_length=unit_length)

  def _remove_proven_p_bubble(self, segment_end1, segment_end2, alternatives,
                              count_tag=None, unit_length=None):
    if count_tag is None:
      count_tag=self.default["count_tag"]
    if unit_length is None:
      unit_length=self.default["unit_length"]
    coverages = [self.try_get_segment(s[0]).coverage(count_tag=count_tag, \
                 unit_length=unit_length) for s in alternatives]
    alternatives.pop(coverages.index(max(coverages)))
    for s in alternatives:
      segment(s[0]).disconnect()



import gfapy
import re

def unsafe_decode(string):
  if string == "*":
    return gfapy.Placeholder()
  else:
    return string

def decode(string):
  obj = unsafe_decode(string)
  validate_decoded(obj)
  return obj

def validate_encoded(string):
  if not re.match(r"^\*$|^[A-Za-z=.]+$", string):
    raise gfapy.FormatError(
      "the string {} is not a valid GFA1 sequence\n".format(repr(string))+
      "(it is not * and does not match the regular expression [A-Za-z=.]+")

def validate_decoded(obj):
  if isinstance(obj, gfapy.Placeholder):
    pass
  elif isinstance(obj, str):
    validate_encoded(obj)
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: str, gfapy.Placeholder)")

def unsafe_encode(obj):
  return str(obj)

def encode(obj):
  if isinstance(obj, gfapy.Placeholder):
    return str(obj)
  elif isinstance(obj, str):
    validate_encoded(obj)
    return obj
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: str, gfapy.Placeholder)")



import gfapy

class References:

  def _compute_required_links(self):
    """
    Computes the list of links which are required to support
    the path.

    Returns
    -------
    list of (gfapy.OrientedLine, gfapy.OrientedLine, gfapy.CIGAR)
      An array, which elements are 3-tuples
      (from oriented segment, to oriented segment, cigar)
    """
    has_undef_overlaps = self._undef_overlaps()
    retval = []
    is_circular = self.is_circular()
    for i in range(len(self.segment_names)):
      j = i+1
      if j == len(self.segment_names):
        if is_circular:
          j = 0
        else:
          break
      if has_undef_overlaps:
        cigar = gfapy.AlignmentPlaceholder()
      else:
        cigar = self.overlaps[i]
      retval.append([self.segment_names[i], self.segment_names[j], cigar])
    return retval

  def _undef_overlaps(self):
    """
    Are the overlaps a single "*"? This is a compact representation of
    a linear path where all CIGARs are "*".

    Returns
    -------
    bool
    """
    return len(self.overlaps) == 1 and gfapy.is_placeholder(self.overlaps[0])

  def _initialize_references(self):
    self._initialize_links()
    self._initialize_segments()

  def _initialize_links(self):
    self._refs["links"] = []
    for from_segment, to_segment, cigar in self._compute_required_links():
      l = None
      orient = "+"
      if self._gfa.segment(from_segment.line) and self._gfa.segment(to_segment.line):
        l = self._gfa._search_link(from_segment, to_segment, cigar)
        if l is not None and l.is_compatible_complement(from_segment, to_segment, cigar):
          orient = "-"
      if l is None:
        if self._gfa._segments_first_order:
          raise gfapy.NotFoundError("Path: {}\n".format(self)+
          "requires a non-existing link:\n"+
          "from={} to={} cigar={}".format(from_segment, to_segment, cigar))
        l = gfapy.line.edge.Link({"from_segment" : from_segment.line,
                                  "from_orient" : from_segment.orient,
                                  "to_segment" : to_segment.line,
                                  "to_orient" : to_segment.orient,
                                  "overlap" : cigar},
                                 virtual = True,
                                 version = "gfa1")
        l.connect(self._gfa)
      self._refs["links"].append(gfapy.OrientedLine(l,orient))
      l._add_reference(self, "paths")

  def _initialize_segments(self):
    for sn_with_o in self.segment_names:
      s = self._gfa.segment(sn_with_o.line)
      sn_with_o.line = s
      s._add_reference(self, "paths")

  def _backreference_keys(self, ref, key_in_ref):
    if ref.record_type == "L":
      return ["links"]
    elif ref.record_type == "S":
      return ["segment_names"]



import unittest
import gfapy

class TestUnitNumericArray(unittest.TestCase):

  def test_integer_type(self):
    v = {b: 2**(b/2) for b in [8,16,32,64,128]}
    self.assertEqual("C", gfapy.NumericArray.integer_type((0,v[8])))
    self.assertEqual("c", gfapy.NumericArray.integer_type((-1,v[8])))
    self.assertEqual("S", gfapy.NumericArray.integer_type((0,v[16])))
    self.assertEqual("s", gfapy.NumericArray.integer_type((-1,v[16])))
    self.assertEqual("I", gfapy.NumericArray.integer_type((0,v[32])))
    self.assertEqual("i", gfapy.NumericArray.integer_type((-1,v[32])))
    self.assertRaises(gfapy.ValueError,
        gfapy.NumericArray.integer_type, (0,v[64]))
    self.assertRaises(gfapy.ValueError,
        gfapy.NumericArray.integer_type, (-1,v[64]))
    self.assertRaises(gfapy.ValueError,
        gfapy.NumericArray.integer_type, (0,v[128]))
    self.assertRaises(gfapy.ValueError,
        gfapy.NumericArray.integer_type, (-1,v[128]))



import gfapy

class Topology:

  def is_cut_link(self, link):
    """Does the removal of a dovetail overlap split a connected component?

    Note:
      only dovetail overlaps are considered as connections

    Parameters:
      link (Line) : an edge instance, which represents a dovetail overlap

    Returns:
       bool
    """
    if link.is_circular():
      return False
    if not link.get("from").dovetails_of_end(\
             gfapy.invert(link.from_end.end_type)):
      return True
    if not link.to.dovetails_of_end(gfapy.invert(link.to_end.end_type)):
      return True
    c = {}
    for et in ["from", "to"]:
      c[et] = set()
      visited = set()
      segend = link.get("from") if et == "from" else link.to
      visited.append(segend.name)
      visited.append(link.other_end(segend).name)
      self.__traverse_component(segend, c[et], visited)
    return c["from"] != c["to"]

  def is_cut_segment(self, segment):
    """Does the removal of a segment split a connected component?

    Note:
      only dovetail overlaps are considered as connections

    Parameters:
      segment (str, Line) : a segment name or instance

    Returns:
       bool
    """
    if isinstance(segment, str):
      segment = self.try_get_segment(segment)
    if segment._connectivity() in [(0,0),(0,1),(1,0)]:
      return False
    start_points = set()
    for et in ["L", "R"]:
      for l in segment.dovetails_of_end(et):
        start_points.append(l.other_end(\
            gfapy.SegmentEnd(segment_name, et)).inverted())
    cc = []
    for start_point in start_points:
      cc.append(set())
      visited = set()
      visited.append(segment_name)
      traverse_component(start_point, cc[-1], visited)
    return any(c != cc[0] for c in cc)

  def segment_connected_component(self, segment, visited = None):
    """Compute the connected component to which a segment belong.

    Note:
      only dovetail overlaps are considered as connections

    Parameters:
      segment (str, Line) : a segment name or instance

    Returns:
       list : a list of segment instances
    """
    if visited is None:
      visited = set()
    if isinstance(segment, gfapy.Line):
      segment_name = segment.name
    else:
      segment_name = segment
      segment = self.segment(segment)
    visited.add(segment_name)
    c = [segment]
    for e in ["L", "R"]:
      self.__traverse_component(gfapy.SegmentEnd(segment, e), c, visited)
    return c

  def connected_components(self):
    """Compute the connected components of the graph.

    Note:
      only dovetail overlaps are considered as connections

    Returns:
       list : a list of lists of segment instances; each sublist is
         a connected component
    """
    components = []
    visited = set()
    for sn in self.segment_names:
      if sn not in visited:
        components.append(self.segment_connected_component(sn, visited))
    return components

  def split_connected_components(self):
    """Split the connected components of the graph.

    Note:
      only dovetail overlaps are considered as connections

    Returns:
      list of Gfa
    """
    retval = []
    for cc in self.connected_components():
      gfa2 = self.clone
      gfa2.rm(gfa2.segment_names - cc)
      retval.append(gfa2)
    return retval

  @property
  def n_dead_ends(self):
    """Number of dead ends in the graph.

    A dead end is a segment end which has no dovetail overlaps.

    Returns:
      int
    """
    n = 0
    for s in self.segments:
      if not s.dovetails_L: n+=1
      if not s.dovetails_R: n+=1
    return n

  @property
  def n_dovetails(self):
    """Number of dovetail overlaps in the graph.

    Returns:
      int
    """
    n = 0
    for s in self.segments:
      n += len(s.dovetails_L)
      n += len(s.dovetails_R)
    return n // 2

  @property
  def n_internals(self):
    """Number of non-dovetail non-containment overlaps in the graph.

    Returns:
      int
    """
    n = 0
    for s in self.segments:
      n += len(s.internals)
    return n // 2

  @property
  def n_containments(self):
    """Number of containment overlaps in the graph.

    Returns:
      int
    """
    n = 0
    for s in self.segments:
      n += len(s.edges_to_contained)
      n += len(s.edges_to_containers)
    return n // 2

  def info(self, short):
    pass

  def __traverse_component(self, segment_end, c, visited):
    s = segment_end.segment
    assert(isinstance(s, gfapy.Line))
    for l in s.dovetails_of_end(segment_end.end_type):
      oe = l.other_end(segment_end)
      sn = oe.name
      s = oe.segment
      if sn in visited:
        continue
      visited.add(sn)
      c.append(s)
      for e in ["L","R"]:
        self.__traverse_component(gfapy.SegmentEnd(s, e), c, visited)



"""
Methods for processing strings as nucleotidic sequences
"""
import gfapy

def rc(sequence, valid = False, rna = False):
  """Compute the reverse complement of a nucleotidic sequence.

  All characters in the IUPAC extended alphabet are supported
  (ACGTUBVHDRYKMSWN). The character ".-=", spaces and newlines
  are left as they are. The case of each character is preserved.

  Returns
    str : reverse complement, without newlines and spaces;
         	"*" if string is "*"

  Parameters:
    sequence (str) : the sequence to reverse-complement
    valid (bool) : if True, the reverse complement of any invalid character
      is the character itself
    rna (bool) : if True, t/T are substituted by u/U in the output

  Raises:
    gfapy.error.ValueError : if valid is False and an invalid character
      (not in the IUPAC extended alphabet for nucleotide sequences, .-=,
      spaces or newline) is found
  """
  if gfapy.is_placeholder(sequence): return sequence
  def fun(c):
    wcc = WCC.get(c, c if valid else None)
    if not wcc:
      raise gfapy.ValueError("{}: no Watson-Crick complement for {}".format(sequence, c))
    return wcc
  retval = "".join(reversed([ fun(c) for c in sequence ]))
  if rna:
    retval = retval.translate(str.maketrans("tT", "uU"))
  return retval

WCC = {"a":"t","t":"a","A":"T","T":"A",
       "c":"g","g":"c","C":"G","G":"C",
       "b":"v","B":"V","v":"b","V":"B",
       "h":"d","H":"D","d":"h","D":"H",
       "R":"Y","Y":"R","r":"y","y":"r",
       "K":"M","M":"K","k":"m","m":"k",
       "S":"S","s":"s","w":"w","W":"W",
       "n":"n","N":"N","u":"a","U":"A",
       "-":"-",".":".","=":"=",
       " ":"","\n":""}
"""Watson-Crick Complements"""

def Sequence(string):
  """Parses the content of a sequence field.

  Parameters:
    string (str) : content of a sequence field

  Returns:
    str, gfapy.Placeholder : if the string is the placeholder
      symbol ``*`` then a placeholder, otherwise the string
      itself
  """
  return gfapy.Placeholder() if (string == "*") else string



import gfapy
import unittest

class TestAPIReferencesVirtual(unittest.TestCase):

  def test_edges_gaps_create_virtual_segments(self):
    data = [
      ["gfa1", {"lines":["L\ta\t+\tb\t-\t*", "C\ta\t-\tb\t+\t100\t*"],
               "m1":"oriented_from", "m2":"oriented_to",
               "sA":"S\ta\t*", "sB":"S\tb\t*",
               "collection":"edges"}],
      ["gfa2", {"lines":["E\t*\ta+\tb-\t0\t100\t900\t1000$\t*"],
               "m1":"sid1", "m2":"sid2",
               "sA":"S\ta\t1000\t*", "sB":"S\tb\t1000\t*",
               "collection":"edges"}],
      ["gfa2", {"lines":["G\t*\ta+\tb-\t1000\t100"],
               "m1":"sid1", "m2":"sid2",
               "sA":"S\ta\t1000\t*", "sB":"S\tb\t1000\t*",
               "collection":"gaps"}]
    ]
    for v,values in data:
      for linestr in values["lines"]:
        g = gfapy.Gfa(version=v)
        line = gfapy.Line(linestr)
        g.append(line)
        self.assertEqual(set(["a", "b"]), set([x.name for x in g.segments]))
        for s in g.segments: assert(s.virtual)
        sA = gfapy.Line(values["sA"])
        g.append(sA)
        self.assertEqual(set(["a", "b"]), set([x.name for x in g.segments]))
        assert(not g.segment("a").virtual)
        assert(g.segment("b").virtual)
        self.assertEqual(sA, getattr(line,values["m1"]).line)
        self.assertEqual(sA, g.segment("a"))
        self.assertEqual([line], getattr(sA,values["collection"]))
        sB = gfapy.Line(values["sB"])
        g.append(sB)
        self.assertEqual(set(["a", "b"]), set([x.name for x in g.segments]))
        assert(not g.segment("b").virtual)
        self.assertEqual(sB, getattr(line,values["m2"]).line)
        self.assertEqual(sB, g.segment("b"))
        self.assertEqual([line], getattr(sB,values["collection"]))

  def test_fragments_create_virtual_segments(self):
    g = gfapy.Gfa(version="gfa2")
    fr = gfapy.Line("F\ta\tread10-\t0\t10\t990\t1000$\t*")
    g.append(fr)
    self.assertEqual(["a"], [x.name for x in g.segments])
    assert(g.segment("a").virtual)
    sA = gfapy.Line("S\ta\t1000\t*")
    g.append(sA)
    self.assertEqual(["a"], [x.name for x in g.segments])
    assert(not g.segment("a").virtual)
    self.assertEqual(sA, fr.sid)
    self.assertEqual(sA, g.segment("a"))
    self.assertEqual([fr], sA.fragments)

  def test_paths_create_virtual_links(self):
    g = gfapy.Gfa(version="gfa1")
    path = gfapy.Line("P\tp1\tb+,ccc-,e+\t10M1I2M,15M")
    g.append(path)
    for i in path.segment_names: assert(i.line.virtual)
    self.assertEqual(set(["b", "ccc", "e"]), set([x.name for x in g.segments]))
    sB = gfapy.Line("S\tb\t*")
    g.append(sB)
    assert(not path.segment_names[0].line.virtual)
    self.assertEqual(sB, path.segment_names[0].line)
    self.assertEqual([path], sB.paths)
    for i in path.links: assert(i.line.virtual)
    l = gfapy.Line("L\tccc\t+\tb\t-\t2M1D10M")
    g.append(l)
    assert(not path.links[0].line.virtual)
    self.assertEqual(l, path.links[0].line)
    self.assertEqual([path], l.paths)
    l = gfapy.Line("L\tccc\t-\te\t+\t15M")
    g.append(l)
    assert(not path.links[1].line.virtual)
    self.assertEqual(l, path.links[1].line)
    self.assertEqual([path], l.paths)

  def test_ordered_groups_create_virtual_unknown_records(self):
    g = gfapy.Gfa(version="gfa2")
    path = gfapy.Line("O\tp1\tchildpath- b+ c- edge-")
    g.append(path)
    for i in path.items:
      assert(i.line.virtual)
      self.assertEqual("\n", i.line.record_type)
    childpath = gfapy.Line("O\tchildpath\tf+ a+")
    g.append(childpath)
    assert(not path.items[0].line.virtual)
    self.assertEqual(childpath, path.items[0].line)
    self.assertEqual([path], childpath.paths)
    sB = gfapy.Line("S\tb\t1000\t*")
    g.append(sB)
    assert(not path.items[1].line.virtual)
    self.assertEqual(sB, path.items[1].line)
    self.assertEqual([path], sB.paths)
    edge = gfapy.Line("E\tedge\te-\tc+\t0\t100\t900\t1000$\t*")
    g.append(edge)
    assert(not path.items[-1].line.virtual)
    self.assertEqual(edge, path.items[-1].line)
    self.assertEqual([path], edge.paths)

  def test_unordered_groups_create_virtual_unknown_records(self):
    g = gfapy.Gfa(version="gfa2")
    set = gfapy.Line("U\tset\tchildpath b childset edge")
    g.append(set)
    for i in set.items:
      assert(i.virtual)
      self.assertEqual("\n", i.record_type)
    childpath = gfapy.Line("O\tchildpath\tf+ a+")
    g.append(childpath)
    assert(not set.items[0].virtual)
    self.assertEqual(childpath, set.items[0])
    self.assertEqual([set], childpath.sets)
    sB = gfapy.Line("S\tb\t1000\t*")
    g.append(sB)
    assert(not set.items[1].virtual)
    self.assertEqual(sB, set.items[1])
    self.assertEqual([set], sB.sets)
    childset = gfapy.Line("U\tchildset\tg edge2")
    g.append(childset)
    assert(not set.items[2].virtual)
    self.assertEqual(childset, set.items[2])
    self.assertEqual([set], childset.sets)
    edge = gfapy.Line("E\tedge\te-\tc+\t0\t100\t900\t1000$\t*")
    g.append(edge)
    assert(not set.items[3].virtual)
    self.assertEqual(edge, set.items[3])
    self.assertEqual([set], edge.sets)




import unittest
import gfapy

class TestNumericArray(unittest.TestCase):

  def test_numeric_arrays(self):
    # creation: new, from array, from string
    a = gfapy.NumericArray([1,2,3,4,5])
    b = gfapy.NumericArray.from_string("i,1,2,3,4,5")
    self.assertEqual(a, b)
    # validation
    a.validate
    gfapy.NumericArray([1,2,3,4,356]).validate
    self.assertRaises(gfapy.ValueError,
        gfapy.NumericArray([1,2.0,3,4,356]).validate)
    self.assertRaises(gfapy.ValueError,
        gfapy.NumericArray([1.0,2.0,3.0,4.0,356]).validate)
    self.assertRaises(gfapy.ValueError,
        gfapy.NumericArray([1,"x",3,4,356]).validate)
    self.assertRaises(gfapy.ValueError,
        gfapy.NumericArray.from_string, "i,1,X,2")
    self.assertRaises(gfapy.FormatError,
        gfapy.NumericArray.from_string, "")
    self.assertRaises(gfapy.FormatError,
        gfapy.NumericArray.from_string, "i,1,2,")
    self.assertRaises(gfapy.TypeError,
        gfapy.NumericArray.from_string, "x,1,2")
    # to string
    a = gfapy.NumericArray([18, 72, 244, 70, 96, 38, 31])
    self.assertEqual("C", a.compute_subtype())
    self.assertEqual("C,18,72,244,70,96,38,31", str(a))
    a[2] = -2
    self.assertEqual("c", a.compute_subtype())
    self.assertEqual("c,18,72,-2,70,96,38,31", str(a))
    a[2] = 280
    self.assertEqual("S", a.compute_subtype())
    self.assertEqual("S,18,72,280,70,96,38,31", str(a))
    a[2] = -280
    self.assertEqual("s", a.compute_subtype())
    self.assertEqual("s,18,72,-280,70,96,38,31", str(a))
    a[2] = 280000
    self.assertEqual("I", a.compute_subtype())
    self.assertEqual("I,18,72,280000,70,96,38,31", str(a))
    a[2] = -280000
    self.assertEqual("i", a.compute_subtype())
    self.assertEqual("i,18,72,-280000,70,96,38,31", str(a))
    a = gfapy.NumericArray([18.0, 72.0, -280000.0, 70.0, 96.0, 38.0, 31.0])
    self.assertEqual("f", a.compute_subtype())
    self.assertEqual("f,18.0,72.0,-280000.0,70.0,96.0,38.0,31.0", str(a))
